# -*- coding: utf-8 -*-
"""
Created on Mon Sep 28 11:35:57 2015

@author: Anton Ievlev
"""

from __future__ import division, print_function, absolute_import, unicode_literals
import numpy as np
from numpy import exp, abs, sqrt, sum, real, imag, arctan2, append


def SHOfunc(parms, w_vec):
    """
    Generates the SHO response over the given frequency band

    Parameters
    -----------
    parms : list or tuple
        SHO parae=(A,w0,Q,phi)
    w_vec : 1D numpy array
        Vector of frequency values
    """
    return parms[0] * exp(1j * parms[3]) * parms[1] ** 2 / \
        (w_vec ** 2 - 1j * w_vec * parms[1] / parms[2] - parms[1] ** 2)


def SHOestimateGuess(w_vec, resp_vec, num_points=5):
    """
    Generates good initial guesses for fitting

    Parameters
    ------------
    w_vec : 1D numpy array or list
        Vector of BE frequencies
    resp_vec : 1D complex numpy array or list
        BE response vector as a function of frequency
    num_points : (Optional) unsigned int
        Quality factor of the SHO peak

    Returns
    ---------
    retval : tuple
        SHO fit parameters arranged as amplitude, frequency, quality factor, phase
    """
    
    ii = np.argsort(abs(resp_vec))[::-1]
    
    a_mat = np.array([])
    e_vec = np.array([])

    for c1 in range(num_points):
        for c2 in range(c1+1,num_points):
            w1 = w_vec[ii[c1]]
            w2 = w_vec[ii[c2]]
            X1 = real(resp_vec[ii[c1]])
            X2 = real(resp_vec[ii[c2]])
            Y1 = imag(resp_vec[ii[c1]])
            Y2 = imag(resp_vec[ii[c2]])
            
            denom = (w1*(X1**2 - X1*X2 + Y1*(Y1 - Y2)) + w2*(-X1*X2 + X2**2 - Y1*Y2 + Y2**2))
            if denom > 0:
                a = ((w1**2 - w2**2)*(w1*X2*(X1**2 + Y1**2) - w2*X1*(X2**2 + Y2**2)))/denom
                b = ((w1**2 - w2**2)*(w1*Y2*(X1**2 + Y1**2) - w2*Y1*(X2**2 + Y2**2)))/denom
                c = ((w1**2 - w2**2)*(X2*Y1 - X1*Y2))/denom
                d = (w1**3*(X1**2 + Y1**2) - w1**2*w2*(X1*X2 + Y1*Y2) - w1*w2**2*(X1*X2 + Y1*Y2) + w2**3*(X2**2 + Y2**2))/denom              
    
                if d>0:
                    a_mat = append(a_mat, [a, b, c, d])
                    
                    A_fit=abs(a+1j*b)/d
                    w0_fit=sqrt(d)
                    Q_fit =-sqrt(d)/c
                    phi_fit=arctan2(-b,-a)
                    
                    H_fit = A_fit*w0_fit**2*exp(1j*phi_fit)/(w_vec ** 2 - 1j * w_vec * w0_fit / Q_fit - w0_fit ** 2)
            
                    e_vec=append(e_vec, sum((real(H_fit) - real(resp_vec)) ** 2) + sum((imag(H_fit) - imag(resp_vec)) ** 2))
    if a_mat.size>0:                
        a_mat=a_mat.reshape(-1,4)        
        
        weight_vec = (1/e_vec)**4
        w_sum=sum(weight_vec)
     
        a_w = sum(weight_vec*a_mat[:,0])/w_sum
        b_w = sum(weight_vec*a_mat[:,1])/w_sum
        c_w = sum(weight_vec*a_mat[:,2])/w_sum
        d_w = sum(weight_vec*a_mat[:,3])/w_sum
         
        A_fit = abs(a_w+1j*b_w)/d_w
        w0_fit = sqrt(d_w)
        Q_fit = -sqrt(d_w)/c_w
        phi_fit = np.arctan2(-b_w,-a_w)
        
        H_fit = A_fit*w0_fit**2*exp(1j*phi_fit)/(w_vec ** 2 - 1j * w_vec * w0_fit / Q_fit - w0_fit ** 2)
    
        if np.std(abs(resp_vec))/np.std(abs(resp_vec-H_fit))<1.2 or w0_fit<np.min(w_vec) or w0_fit>np.max(w_vec):
            p0=SHOfastGuess(w_vec, resp_vec)
        else:           
            p0=np.array([A_fit, w0_fit, Q_fit, phi_fit])
    else:
        p0=SHOfastGuess(w_vec, resp_vec)
        
    return p0


def SHOfastGuess(w_vec, resp_vec, qual_factor=200):
    """
    Default SHO guess from the maximum value of the response

    Parameters
    ------------
    w_vec : 1D numpy array or list
        Vector of BE frequencies
    resp_vec : 1D complex numpy array or list
        BE response vector as a function of frequency
    qual_factor : float
        Quality factor of the SHO peak

    Returns
    retval : 1D numpy array
    ---------
        SHO fit parameters arranged as [amplitude, frequency, quality factor, phase]
    """
    amp_vec = abs(resp_vec)
    i_max = int(len(resp_vec)/2)
    return np.array([np.mean(amp_vec) / qual_factor, w_vec[i_max], qual_factor, np.angle(resp_vec[i_max])])

def SHOlowerBound(w_vec):
    """
    Provides the lower bound for the SHO fitting function

    Parameters
    ----------
    w_vec : 1D numpy array or list
        Vector of BE frequencies

    Returns
    -------
    retval : tuple
        SHO fit parameters arranged as amplitude, frequency, quality factor, phase
    """
    return 0, np.min(w_vec), -1e5, -np.pi


def SHOupperBound(w_vec):
    """
    Provides the upper bound for the SHO fitting function

    Parameters
    ----------
    w_vec: 1D numpy array or list
        Vector of BE frequencies

    Returns
    -------
    retval : tuple
        SHO fit parameters arranged as amplitude, frequency, quality factor, phase
    """
    return 1e5, np.max(w_vec), 1e5, np.pi



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 20 17:42:41 2015

@author: Suhas Somnath
"""
###############################################################################

from __future__ import division, print_function, absolute_import
import numpy as np  # for all array, data operations
import matplotlib.pyplot as plt  # for all plots
from scipy.special import erf
from warnings import warn


def get_fft_stack(image_stack):
    """
    Gets the 2D FFT for a stack of images by applying a blackman window

    Parameters
    ----------
    image_stack : 2D or 3D real numpy array
        Either a 2D matrix [x, y] or a stack of 2D images arranged as [x, y, z or spectral]

    Returns
    -------
    fft_stack : 2D or 3D real numpy array
        2 or 3 dimensional matrix arranged as [x, y, spectral dimension]

    """
    image_stack = np.atleast_3d(image_stack)
    blackman_2d = np.atleast_2d(np.blackman(image_stack.shape[1])) * np.atleast_2d(np.blackman(image_stack.shape[0])).T
    image_stack = np.transpose(image_stack, [2, 0, 1])
    blackman_3d = np.transpose(np.atleast_3d(blackman_2d), [2, 0, 1])
    fft_stack = blackman_3d * image_stack
    fft_stack = np.abs(np.fft.fftshift(np.fft.fft2(fft_stack, axes=(1, 2)), axes=(1, 2)))
    fft_stack = np.transpose(fft_stack, [1, 2, 0])
    return np.squeeze(fft_stack)


def getNoiseFloor(fft_data, tolerance):
    """
    Calculate the noise floor from the FFT data. Algorithm originally written by Mahmut Okatan Baris

    Parameters
    ----------
    fft_data : 1D or 2D complex numpy array
        Signal in frequency space (ie - after FFT shifting) arranged as (channel or repetition, signal)
    tolerance : unsigned float
        Tolerance to noise. A smaller value gets rid of more noise.
        
    Returns
    -------
    noise_floor : 1D real numpy array 
        One value per channel / repetition

    """

    fft_data = np.atleast_2d(fft_data)
    # Noise calculated on the second axis

    noise_floor = np.zeros(fft_data.shape[0])

    for chan in range(fft_data.shape[0]):

        amp = np.abs(fft_data[chan, :])
        num_pts = amp.size
        temp = np.sqrt(np.sum(amp ** 2) / (2 * num_pts))
        threshold = np.sqrt((2 * temp ** 2) * (-np.log(tolerance)))

        bdiff = 1
        b_vec = list([temp])
        # b_vec.append(temp)
        n_b_vec = 1

        while (bdiff > 10 ** -2) and n_b_vec < 50:
            amp[amp > threshold] = 0
            temp = np.sqrt(np.sum(amp ** 2) / (2 * num_pts))
            b_vec.append(temp)
            bdiff = abs(b_vec[1] - b_vec[0])
            threshold = np.sqrt((2 * temp ** 2) * (-np.log(tolerance)))
            del b_vec[0]
            n_b_vec += 1

        noise_floor[chan] = threshold

    return noise_floor


###############################################################################


def downSample(fft_vec, freq_ratio):
    """
    Downsamples the provided data vector
    
    Parameters
    ----------
    fft_vec : 1D complex numpy array
        Waveform that is already FFT shifted
    freq_ratio : float
        new sampling rate / old sampling rate (less than 1)
    
    Returns
    -------
    fft_vec : 1D numpy array
        downsampled waveform

    """
    if freq_ratio >= 1:
        warn('Error at downSample: New sampling rate > old sampling rate')
        return fft_vec

    vec_len = len(fft_vec)
    ind = np.round(float(vec_len) * (0.5 * float(freq_ratio)))
    fft_vec = fft_vec[max(0.5 * vec_len - ind, 0):min(0.5 * vec_len + ind, vec_len)]
    fft_vec = fft_vec * freq_ratio * 2

    return np.fft.ifft(np.fft.ifftshift(fft_vec))


###############################################################################


def noiseBandFilter(num_pts, samp_rate, freqs, freq_widths, show_plots=False):
    """
    Builds a filter that removes specified noise frequencies
    
    Parameters
    ----------
    num_pts : unsigned int
        Number of points in the FFT signal
    samp_rate : unsigned int
        sampling rate    
    freqs : 1D array or list
        Target frequencies as unsigned ints    
    freq_widths : 1D array or list
        Width around the target frequency that should be set to 0\n    
    Note: sampRate, freqs, freq_widths have same units - eg MHz
    
    Returns
    -------
    noise_filter : 1D numpy array
        Array of ones set to 0 at noise bands

    """
    num_pts = abs(int(num_pts))

    w_vec = 1

    # Making code a little more robust with handling different inputs:
    samp_rate = float(samp_rate)
    freqs = np.array(freqs)
    freq_widths = np.array(freq_widths)
    if freqs.ndim != freq_widths.ndim:
        warn('Error in noiseBandFilter: dimensionality of frequencies and frequency widths do not match!')
        return 1
    if freqs.shape != freq_widths.shape:
        warn('Error in noiseBandFilter: shape of frequencies and frequency widths do not match!')
        return 1

    cent = int(round(0.5 * num_pts))

    noise_filter = np.ones(num_pts, dtype=np.int16)

    if show_plots:
        w_vec = np.arange(-0.5 * samp_rate, 0.5 * samp_rate, samp_rate / num_pts)
        fig, ax = plt.subplots(2, 1)
        ax[0].plot(w_vec, noise_filter)
        ax[0].set_yscale('log')
        ax[0].axis('tight')
        ax[0].set_xlabel('Freq')
        ax[0].set_title('Before clean up')

    # Setting noise freq bands to 0
    for cur_freq, d_freq in zip(freqs, freq_widths):
        ind = int(round(num_pts * (cur_freq / samp_rate)))
        sz = int(round(cent * d_freq / samp_rate))
        noise_filter[cent - ind - sz:cent - ind + sz + 1] = 0
        noise_filter[cent + ind - sz:cent + ind + sz + 1] = 0

    if show_plots:
        ax[1].plot(w_vec, noise_filter)
        ax[1].set_yscale('log')
        ax[1].axis('tight')
        ax[1].set_xlabel('Freq')
        ax[1].set_title('After clean up')
        plt.show()

    return noise_filter


###############################################################################


def makeLPF(num_pts, samp_rate, f_cutoff, roll_off=0.05):
    """
    Builds a low pass filter
    
    Parameters
    ----------
    num_pts : unsigned int
        Points in the FFT. Assuming Signal in frequency space (ie - after FFT shifting)
    samp_rate : unsigned integer
        Sampling rate
    f_cutoff : unsigned integer
        Cutoff frequency for filter
    roll_off : 0 < float < 1
        Frequency band over which the filter rolls off. rol off = 0.05 on a 
        100 kHz low pass filter -> roll off from 95 kHz (1) to 100 kHz (0)
        
    Returns
    -------
    LPF : 1D numpy array describing the low pass filter

    """

    num_pts = abs(int(num_pts))

    cent = int(round(0.5 * num_pts))

    if f_cutoff >= 0.5 * samp_rate:
        warn('Error in LPFClip --> LPF too high! Skipping')
        return 1

    # BW = 0.1; %MHz - Nothing beyond BW.
    roll_off *= f_cutoff  # MHz

    sz = int(np.round(num_pts * (roll_off / samp_rate)))
    ind = int(np.round(num_pts * (f_cutoff / samp_rate)))

    lpf = np.zeros(num_pts, dtype=np.float32)

    extent = 5.0
    t2 = np.linspace(-extent / 2, extent / 2, num=sz)
    smoothing = 0.5 * (1 + erf(t2))

    lpf[cent - ind:cent - ind + sz] = smoothing
    lpf[cent - ind + sz:cent + ind - sz + 1] = 1
    lpf[cent + ind - sz + 1:cent + ind + 1] = 1 - smoothing

    return lpf  # return the filter itself so that it may be repeatedly used outside


###############################################################################


def harmonicsPassFilter(num_pts, samp_rate, first_freq, band_width, num_harm, do_plots=False):
    """
    Builds a filter that only keeps N harmonics
    
    Parameters
    ----------
    num_pts : unsigned int
        Number of points in the FFt signal
    samp_rate : unsigned int
        Sampling rate
    first_freq : unsigned int
        Frequency of the first harmonic
    band_width : unsigned int
        Frequency band around each harmonic that needs to be preserved
    num_harm : unsigned int
        Number of harmonics to preserve
    do_plots : Boolean (optional)
        Whether or not to generate plots. Not necessary after debugging

    Note that the frequency values must all have the same units
    
    Returns
    -------
    harm_filter : 1D numpy array
        0s where the signal is to be rejected and 1s at harmonics
        
    """

    num_pts = abs(int(num_pts))

    harm_filter = np.ones(num_pts, dtype=np.int16)

    cent = int(round(0.5 * num_pts))

    w_vec = 1

    if do_plots:
        print(
        'OnlyKeepHarmonics: samp_rate = %2.1e Hz, first harmonic = %3.2f Hz, %d harmonics w/- %3.2f Hz bands\n' % (
        samp_rate, first_freq, num_harm, band_width))
        w_vec = np.arange(-samp_rate / 2.0, samp_rate / 2.0, samp_rate / num_pts)
        fig, ax = plt.subplots(figsize=(5, 5))
        ax.plot(w_vec, harm_filter)
        ax.set_title('Raw')

    sz = int(round(cent * band_width / samp_rate))

    # First harmonic
    ind = int(round(num_pts * (first_freq / samp_rate)))
    if ind >= num_pts:
        warn('Invalid harmonic frequency')
        return 1

    harm_filter[max(cent - ind + sz + 1, 0):min(num_pts, cent + ind - sz)] = 0

    if do_plots:
        fig2, ax2 = plt.subplots(figsize=(5, 5))
        ax2.plot(w_vec, harm_filter)
        ax2.set_title('Step 1')

    # Last harmonic
    ind = int(round(num_pts * (num_harm * first_freq / samp_rate)))
    harm_filter[:cent - ind - sz] = 0
    harm_filter[cent + ind + sz + 1:] = 0

    if do_plots:
        fig3, ax3 = plt.subplots(figsize=(5, 5))
        ax3.plot(w_vec, harm_filter)
        ax3.set_title('Step 2')

    if num_harm == 1:
        return harm_filter

    for harm_ind in range(1, num_harm):
        ind = int(round(num_pts * (harm_ind * first_freq / samp_rate)))
        ind2 = int(round(num_pts * ((harm_ind + 1) * first_freq / samp_rate)))
        harm_filter[cent - ind2 + sz + 1:cent - ind - sz] = 0
        harm_filter[cent + ind + sz + 1:cent + ind2 - sz] = 0
        if do_plots:
            fig4, ax4 = plt.subplots(figsize=(5, 5))
            ax4.plot(w_vec, harm_filter)
            ax4.set_title('Step %d' % (harm_ind + 2))

    return harm_filter


    ###############################################################################

    # def removeNoiseHarmonics(F_AI_vec,samp_rate,noise_combs):
    #     """
    #     Removes specified noise frequencies from the signal
    #
    #     Parameters
    #     ---------------------
    #     * F_AI_vec -- matrix (chan x pts) already FFT + FFT shifted
    #
    #     * sampRate -- sampling rate
    #
    #     * freqs -- 1D array of target frequencies
    #
    #     * freqWidths -- 1D array of frequency windows that correspond to freqs that should be set to 0
    #
    #     * Note: sampRate, freqs, freqWidths have same units - eg MHz
    #     """
    #     numpts = size(F_AI_vec,2);
    #     freqs = []; freqWidths = [];
    #     for k1 = 1:length(noiseCombs):
    #         stFreq = noiseCombs(k1).first_harm_freq;
    #         nBands = noiseCombs(k1).num_harmonics;
    #         if nBands < 0
    #             nBands =
    #         end
    #     end
    #     F_AI_vec = removeNoiseFreqs(F_AI_vec,sampRate,freqs,freqWidths);



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 04 11:12:45 2016

@author: Suhas Somnath
"""

from __future__ import division, print_function, absolute_import, unicode_literals

from os import path, remove  # File Path formatting

import numpy as np  # For array operations
from scipy.io.matlab import loadmat  # To load parameters stored in Matlab .mat file

from .df_utils.gmode_utils import readGmodeParms
from .translator import Translator  # Because this class extends the abstract Translator class
from .utils import make_position_mat, get_position_slicing, generate_dummy_main_parms
from ..hdf_utils import getH5DsetRefs, linkRefs
from ..io_hdf5 import ioHDF5  # Now the translator is responsible for writing the data.
from ..microdata import MicroDataGroup, MicroDataset  # The building blocks for defining heirarchical storage in the H5 file


class GDMTranslator(Translator):
    """
    Translates G-mode w^2 datasets from .mat files to .h5
    """
    
    def translate(self, parm_path):
        """
        Basic method that translates .mat data files to a single .h5 file
        
        Parameters
        ------------
        parm_path : string / unicode
            Absolute file path of the parameters .mat file. 
            
        Returns
        ----------
        h5_path : string / unicode
            Absolute path of the translated h5 file
        """
        (folder_path, file_name) = path.split(parm_path)
        (file_name, base_name) = path.split(folder_path)
        h5_path = path.join(folder_path,base_name+'.h5')
        
        # Read parameters
        parm_dict = readGmodeParms(parm_path)
        
        # Add the w^2 specific parameters to this list
        parm_data = loadmat(parm_path, squeeze_me=True, struct_as_record=True)
        freq_sweep_parms = parm_data['freqSweepParms']
        parm_dict['freq_sweep_delay'] = np.float(freq_sweep_parms['delay'].item())
        gen_sig = parm_data['genSig']
        parm_dict['wfm_fix_d_fast'] = np.int32(gen_sig['restrictT'].item())
        freq_array = np.float32(parm_data['freqArray'])
        
        # prepare and write spectroscopic values
        samp_rate = parm_dict['IO_down_samp_rate_[Hz]']
        num_bins = int(parm_dict['wfm_n_cycles']*parm_dict['wfm_p_slow']*samp_rate)
        
        w_vec = np.arange(-0.5*samp_rate, 0.5*samp_rate,np.float32(samp_rate/num_bins))
        
        # There is most likely a more elegant solution to this but I don't have the time... Maybe np.meshgrid
        spec_val_mat = np.zeros((len(freq_array)*num_bins,2), dtype=np.float32)
        spec_val_mat[:,0] = np.tile(w_vec,len(freq_array)) 
        spec_val_mat[:,1] = np.repeat(freq_array,num_bins) 
        
        spec_ind_mat = np.zeros((2,len(freq_array)*num_bins), dtype=np.int32)
        spec_ind_mat[0,:] = np.tile(np.arange(num_bins),len(freq_array)) 
        spec_ind_mat[1,:] = np.repeat(np.arange(len(freq_array)),num_bins) 
        
        num_rows = parm_dict['grid_num_rows']
        num_cols = parm_dict['grid_num_cols']
        parm_dict['data_type'] = 'GmodeW2'
            
        num_pix = num_rows*num_cols
        
        pos_mat = make_position_mat([num_cols, num_rows])
        pos_slices = get_position_slicing(['X', 'Y'], num_pix)
        
        # Now start creating datasets and populating:
        ds_pos_ind = MicroDataset('Position_Indices', np.uint32(pos_mat))
        ds_pos_ind.attrs['labels'] = pos_slices
        ds_pos_val = MicroDataset('Position_Values', np.float32(pos_mat))
        ds_pos_val.attrs['labels'] = pos_slices
              
        ds_spec_inds = MicroDataset('Spectroscopic_Indices', np.uint32(spec_ind_mat))
        ds_spec_inds.attrs['labels'] = {'Response Bin Index':(slice(0,1), slice(None)), 'Excitation Frequency Index':(slice(1,2),slice(None))}
        
        ds_spec_vals = MicroDataset('Spectroscopic_Values', spec_val_mat)
        ds_spec_vals.attrs['labels'] = {'Response Bin':(slice(0,1), slice(None)), 'Excitation Frequency':(slice(1,2), slice(None))}
                
        ds_ex_freqs = MicroDataset('Excitation_Frequencies', freq_array)
        ds_bin_freq = MicroDataset('Bin_Frequencies', w_vec)
        
        # Minimize file size to the extent possible.
        # DAQs are rated at 16 bit so float16 should be most appropriate.
        # For some reason, compression is more effective on time series data
        ds_main_data = MicroDataset('Raw_Data', data=[], maxshape=(num_pix, len(freq_array)*num_bins), dtype=np.float32, chunking=(1,num_bins), compression='gzip')
        
        chan_grp = MicroDataGroup('Channel_000')		
        chan_grp.attrs = parm_dict        
        chan_grp.addChildren([ds_pos_ind, ds_pos_val, ds_spec_inds, ds_spec_vals, 
								ds_ex_freqs, ds_bin_freq, ds_main_data])
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.addChildren([chan_grp])
        
        spm_data = MicroDataGroup('')
        global_parms = generate_dummy_main_parms()
        global_parms['grid_size_x'] = parm_dict['grid_num_cols'];
        global_parms['grid_size_y'] = parm_dict['grid_num_rows'];
        # assuming that the experiment was completed:        
        global_parms['current_position_x'] = parm_dict['grid_num_cols']-1;
        global_parms['current_position_y'] = parm_dict['grid_num_rows']-1;
        global_parms['data_type'] = parm_dict['data_type'] #self.__class__.__name__
        global_parms['translator'] = 'W2'
        spm_data.attrs = global_parms
        spm_data.addChildren([meas_grp])
        
        if path.exists(h5_path):
            remove(h5_path)
        
        # Write everything except for the main data.
        hdf = ioHDF5(h5_path)
        
        h5_refs = hdf.writeData(spm_data)
                    
        h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
            
        #Now doing linkrefs:
        aux_ds_names = ['Position_Indices','Position_Values',
                     'Spectroscopic_Indices','Spectroscopic_Values',
                     'Excitation_Frequencies', 'Bin_Frequencies']
        linkRefs(h5_main, getH5DsetRefs(aux_ds_names, h5_refs))

        # Now read the raw data files:
        pos_ind = 0
        for row_ind in range(1,num_rows+1):
            for col_ind in range(1,num_cols+1):
                file_path = path.join(folder_path,'fSweep_r'+str(row_ind)+'_c'+str(col_ind)+'.mat')
                print('Working on row {} col {}'.format(row_ind,col_ind))
                if path.exists(file_path):
                    # Load data file
                    pix_data = loadmat(file_path, squeeze_me=True)
                    pix_mat = pix_data['AI_mat']
                    # Take the inverse FFT on 2nd dimension
                    pix_mat = np.fft.ifft(np.fft.ifftshift(pix_mat,axes=1), axis=1)
                    # Verified with Matlab - no conjugate required here.
                    pix_vec = pix_mat.transpose().reshape(pix_mat.size)
                    h5_main[pos_ind,:] = np.float32(pix_vec)
                    hdf.flush() # flush from memory!
                else:
                    print('File not found for: row {} col {}'.format(row_ind,col_ind))
                pos_ind +=1
                if (100.0*(pos_ind)/num_pix)%10 == 0:
                    print('completed translating {} %'.format(int(100*pos_ind/num_pix)))
                    
        hdf.close()
        
        return h5_path


"""
Created on 12/15/16 10:08 AM
@author: Numan Laanait
"""

from __future__ import division, print_function, absolute_import, unicode_literals
from warnings import warn
import numpy as np
import sys
import multiprocessing as mp
from .guess_methods import GuessMethods
from .fit_methods import Fit_Methods
import scipy


def targetFuncGuess(args, **kwargs):
    """
    This is just creates mappable function for multiprocessing guess
    :param args:
    :param kwargs:
    :return:
    """
    opt = args[-1]
    func = opt._guessFunc()
    results = func(args[0])
    return results


def targetFuncFit(args, **kwargs):
    """
    Needed to create mappable function for multiprocessing
    :param args:
    :param kwargs:
    :return:
    """
    opt = args[-1]
    solver, solver_options, func = opt._initiateSolverAndObjFunc()
    results = solver(func, args[1], args=[args[0]])
    return results


class Optimize(object):
    """
    In charge of all optimization and computation and is used within the Model Class.
    """

    def __init__(self, data=np.array([]), guess=np.array([]), parallel=True):
        """

        :param data:
        """
        if isinstance(data, np.ndarray):
            self.data = data
        if isinstance(guess, np.ndarray):
            self.guess = guess
        else:
            warn('Error: data and guess must be numpy.ndarray. Exiting...')
            sys.exit()
        self._parallel = parallel

    def _guessFunc(self):
        gm = GuessMethods()
        if self.strategy in gm.methods:
            func = gm.__getattribute__(self.strategy)(**self.options)
            return func
        else:
            warn('Error: %s is not implemented in pycroscopy.analysis.GuessMethods to find guesses' % self.strategy)

    def computeGuess(self, processors=1, strategy='wavelet_peaks',
                     options={"peak_widths": np.array([10, 200]), "peak_step": 20}, **kwargs):
        """
        Computes the guess function using numerous cores

        Parameters
        ----------
        processors : unsigned int
            Number of logical cores to use for computing
        strategy: string
            Default is 'Wavelet_Peaks'.
            Can be one of ['wavelet_peaks', 'relative_maximum', 'gaussian_processes']. For updated list, run GuessMethods.methods
        options: dict
            Default: Options for wavelet_peaks{"peaks_widths": np.array([10,200]), "peak_step":20}.
            Dictionary of options passed to strategy. For more info see GuessMethods documentation.

        kwargs:
            processors: int
                number of processors to use. Default all processors on the system except for 1.

        Returns
        -------
        results : unknown
            unknown
        """
        self.strategy = strategy
        self.options = options
        gm = GuessMethods()
        if strategy in gm.methods:
            # func = gm.__getattribute__(strategy)(**options)
            results = list()
            if self._parallel:
                # start pool of workers
                print('Computing Jobs In parallel ... launching %i kernels...' % processors)
                pool = mp.Pool(processors)
                # Vectorize tasks
                tasks = [(vector, self) for vector in self.data]
                chunk = int(self.data.shape[0] / processors)
                # Map them across processors
                jobs = pool.imap(targetFuncGuess, tasks, chunksize=chunk)
                # get Results from different processes
                results = [j for j in jobs]
                print('Extracted Results...')
                # Finished reading the entire data set
                print('closing %i kernels...' % processors)
                pool.close()
                return results

            else:
                print("Computing Guesses In Serial ...")
                results = [targetFuncGuess((vector, self)) for vector in self.data]
                return results
        else:
            warn('Error: %s is not implemented in pycroscopy.analysis.GuessMethods to find guesses' % strategy)

    def _initiateSolverAndObjFunc(self):
        if self.solver_type in scipy.optimize.__dict__.keys():
            solver = scipy.optimize.__dict__[self.solver_type]
        if self.obj_func is None:
            fm = Fit_Methods()
            func = fm.__getattribute__(self.obj_func_name)(self.obj_func_xvals)
        return solver, self.solver_options, func

    def computeFit(self, processors=1, solver_type='least_squares', solver_options={},
                   obj_func={'class': 'Fit_Methods', 'obj_func': 'SHO', 'xvals': np.array([])}):
        """

        Parameters
        ----------
        processors : unsigned int
            Number of logical cores to use for computing
        solver_type : string
            Optimization solver to use (minimize,least_sq, etc...). For additional info see scipy.optimize
        solver_options: dict()
            Default: dict()
            Dictionary of options passed to solver. For additional info see scipy.optimize
        obj_func: dict()
            Default is 'SHO'.
            Can be one of ['wavelet_peaks', 'relative_maximum', 'gaussian_processes'].
            For updated list, run GuessMethods.methods

        Returns
        -------
        results : unknown
            unknown
        """
        self.solver_type = solver_type
        self.solver_options = solver_options
        if self.solver_type not in scipy.optimize.__dict__.keys():
            warn('Solver %s does not exist!. For additional info see scipy.optimize' % (solver_type))
            sys.exit()
        if obj_func['class'] is None:
            self.obj_func = obj_func['obj_func']
        else:
            self.obj_func_xvals = obj_func['xvals']
            self.obj_func = None
            self.obj_func_name = obj_func['obj_func']
            self.obj_func_class = obj_func['class']

        if self._parallel:
            # start pool of workers
            print('Computing Jobs In parallel ... launching %i kernels...' % processors)
            pool = mp.Pool(processors)
            # Vectorize tasks
            tasks = [(vector, guess, self) for vector, guess in zip(self.data, self.guess)]
            chunk = int(self.data.shape[0] / processors)
            # Map them across processors
            jobs = pool.imap(targetFuncFit, tasks, chunksize=chunk)
            # get Results from different processes
            results = [j for j in jobs]
            # Finished reading the entire data set
            print('closing %i kernels...' % processors)
            pool.close()
            return results

        else:
            print("Computing Fits In Serial ...")
            tasks = [(vector, guess, self) for vector, guess in zip(self.data, self.guess)]
            results = [targetFuncFit(task) for task in tasks]
            return results



'''
Created on Aug 7, 2015

@author: James Anderson

link: https://github.com/jamesra/dm4reader
'''

from __future__ import division, print_function, absolute_import, unicode_literals
import collections
import struct
import array 

DM4Header = collections.namedtuple('DM4Header', ('version', 'root_length', 'little_endian')) 
DM4TagHeader = collections.namedtuple('DM4TagHeader', ('type', 'name', 'byte_length', 'array_length', 'data_type_code', 'header_offset', 'data_offset'))
DM4DirHeader = collections.namedtuple('DM4DirHeader', ('type', 'name', 'byte_length', 'sorted', 'closed', 'num_tags', 'data_offset'))

DM4Tag = collections.namedtuple('DM4Tag', ('name', 'data_type_code', 'data'))


DM4DataType = collections.namedtuple('DM4DataTypes', ('num_bytes', 'signed', 'type_format'))

DM4DataTypeDict = {2: DM4DataType(2,True,'h'), #2byte signed integer
                   3: DM4DataType(4,True,'i'), #4byte signed integer
                   4: DM4DataType(2,False,'H'), #2byte unsigned integer
                   5: DM4DataType(4,False,'I'), #4byte unsigned integer
                   6: DM4DataType(4,False,'f'), #4byte float
                   7: DM4DataType(8,False,'d'), #8byte float
                   8: DM4DataType(1,False, '?'),
                   9: DM4DataType(1,False, 'c'),
                   10: DM4DataType(1,True,'b'),
                   11: DM4DataType(8,True,'q'),
                   12: DM4DataType(8,True,'Q')
                   }

DM4_header_size = 4+8+4
DM4_root_tag_dir_header_size = 1+1+8


def tag_is_directory(tag):
    return tag.type == 20
 
def read_header_dm4(dmfile):
    dmfile.seek(0)
    version = struct.unpack_from('>I', dmfile.read(4))[0] #int.from_bytes(dmfile.read(4), byteorder='big')
    rootlength = struct.unpack_from('>Q', dmfile.read(8))[0]
    byteorder = struct.unpack_from('>I', dmfile.read(4))[0]
    
    little_endian = byteorder == 1
    
    return DM4Header(version, rootlength, little_endian)

def _get_endian_str(endian):
    '''
    DM4 header encodes little endian as byte value 1 in the header
    :return: 'big' or 'little' for use with python's int.frombytes function
    '''
    if isinstance(endian, str):
        return endian 
    
    assert(isinstance(endian,int))
    if endian == 1:
        return 'little'
    
    return 'big'   

def _get_struct_endian_str(endian):
    '''
    DM4 header encodes little endian as byte value 1 in the header.  However when that convention is followed the wrong values are read.  So this implementation is reversed.
    :return: '>' or '<' for use with python's struct.unpack function
    '''
    if isinstance(endian,str):
        if endian == 'little':
            return '>' #Little Endian
        else:
            return '<' #Big Endian
    else:
        if endian == 1:
            return '>' #Little Endian
        else:
            return '<' #Big Endian

def read_root_tag_dir_header_dm4(dmfile, endian):
    '''Read the root directory information from a dm4 file.
       File seek position is left at end of root_tag_dir_header'''
    if not isinstance(endian, str):
        endian = _get_struct_endian_str(endian)
        
    dmfile.seek(DM4_header_size)
        
    issorted = struct.unpack_from(endian + 'b', dmfile.read(1))[0]
    isclosed = struct.unpack_from(endian + 'b', dmfile.read(1))[0]
    num_tags = struct.unpack_from('>Q', dmfile.read(8))[0] #DM4 specifies this property as always big endian
      
    return DM4DirHeader(20, None, 0, issorted, isclosed, num_tags, DM4_header_size)

def read_tag_header_dm4(dmfile, endian):
    '''Read the tag from the file.  Leaves file at the end of the tag data, ready to read the next tag from the file'''
    tag_header_offset = dmfile.tell()
    tag_type = struct.unpack_from(endian + 'B', dmfile.read(1))[0]
    if tag_type == 20:
        return _read_tag_dir_header_dm4(dmfile, endian)
    if tag_type == 0:
        return None
    
    tag_name = _read_tag_name(dmfile, endian)
    tag_byte_length = struct.unpack_from('>Q', dmfile.read(8))[0]  #DM4 specifies this property as always big endian
    
    tag_data_offset = dmfile.tell()
    
    _read_tag_garbage_str(dmfile)
 
    (tag_array_length,  tag_array_types) = _read_tag_data_info(dmfile)
     
    dmfile.seek(tag_data_offset + tag_byte_length)
    return DM4TagHeader(tag_type, tag_name, tag_byte_length, tag_array_length, tag_array_types[0], tag_header_offset, tag_data_offset)

def _read_tag_name(dmfile, endian):
    tag_name_len = struct.unpack_from('>H',dmfile.read(2))[0] #DM4 specifies this property as always big endian
    tag_name = None
    if tag_name_len > 0:
        data = dmfile.read(tag_name_len)
        try:
            tag_name = data.decode('utf-8', errors='ignore')
        except UnicodeDecodeError as e:
            tag_name = None
            pass
    
    return tag_name
    
def _read_tag_dir_header_dm4(dmfile, endian):

    tag_name = _read_tag_name(dmfile, endian)
    tag_byte_length = struct.unpack_from('>Q', dmfile.read(8))[0] #DM4 specifies this property as always big endian
    issorted = struct.unpack_from(endian + 'b', dmfile.read(1))[0]
    isclosed = struct.unpack_from(endian + 'b', dmfile.read(1))[0]
    num_tags =  struct.unpack_from('>Q', dmfile.read(8))[0] #DM4 specifies this property as always big endian
    
    data_offset = dmfile.tell()
    
    return DM4DirHeader(20, tag_name, tag_byte_length, issorted, isclosed, num_tags, data_offset)

def _read_tag_garbage_str(dmfile):
    '''
    DM4 has four bytes of % symbols in the tag.  Ensure it is there.
    '''
    garbage_str = dmfile.read(4).decode('utf-8')
    assert(garbage_str == '%%%%')
    
def _read_tag_data_info(dmfile):
    tag_array_length = struct.unpack_from('>Q', dmfile.read(8))[0] #DM4 specifies this property as always big endian
    format_str = '>' + tag_array_length * 'q' #Big endian signed long
     
    tag_array_types = struct.unpack_from(format_str, dmfile.read(8*tag_array_length))
    
    return (tag_array_length,  tag_array_types)


def _read_tag_data(dmfile, tag, endian):
    assert(tag.type == 21)
    try:
        
        endian = _get_struct_endian_str(endian)
        dmfile.seek(tag.data_offset)
         
        _read_tag_garbage_str(dmfile)
        (tag_array_length,  tag_array_types) = _read_tag_data_info(dmfile)
        
        tag_data_type_code = tag_array_types[0]
        
        if tag_data_type_code == 15:
            return read_tag_data_group(dmfile, tag, endian)
        elif tag_data_type_code == 20:
            return read_tag_data_array(dmfile, tag, endian)
        
        if not tag_data_type_code in DM4DataTypeDict:
            print ("Missing type " + str(tag_data_type_code))
            return None
        
        return _read_tag_data_value(dmfile, endian, tag_data_type_code)
    
    finally:
        #Ensure we are in the correct position to read the next tag regardless of how reading this tag goes
        dmfile.seek(tag.data_offset + tag.byte_length)
        

def _read_tag_data_value(dmfile, endian, type_code):
    data_type = DM4DataTypeDict[type_code]
    format_str = _get_struct_endian_str(endian) + data_type.type_format
    byte_data = dmfile.read(data_type.num_bytes)
    
    return struct.unpack_from(format_str, byte_data)[0] 

def read_tag_data_group(dmfile, tag, endian):
    endian = _get_struct_endian_str(endian)
    dmfile.seek(tag.data_offset)
    
    _read_tag_garbage_str(dmfile)
    (tag_array_length,  tag_array_types) = _read_tag_data_info(dmfile)
        
    tag_data_type = tag_array_types[0]
    assert(tag_data_type == 15)
    
    length_groupname = tag_array_types[1]
    number_of_entries_in_group = tag_array_types[2]
    field_data = tag_array_types[3:]
    
    field_types_list = []
    
    for iField in range(0,number_of_entries_in_group):
        fieldname_length = field_data[iField*2]
        fieldname_type = field_data[(iField*2)+1]
        field_types_list.append(fieldname_type)
        
    fields_data = []
    for field_type in field_types_list:
        field_data = _read_tag_data_value(dmfile, endian, field_type)
        fields_data.append(field_data)
        
    return fields_data

def read_tag_data_array(dmfile, tag, endian):
    endian = _get_struct_endian_str(endian)
    dmfile.seek(tag.data_offset)
    
    _read_tag_garbage_str(dmfile)
    
    (tag_array_length,  tag_array_types) = _read_tag_data_info(dmfile)
        
    assert(tag_array_types[0] == 20)
    array_data_type_code = tag_array_types[1]
    array_length = tag_array_types[2]
    
    if array_data_type_code == 15:
        return "Array of groups length %d and type %d" % (array_length, array_data_type_code)
     
    assert(len(tag_array_types) == 3)
    
    data_type = DM4DataTypeDict[array_data_type_code]
    
    data = array.array(data_type.type_format)
    data.fromfile(dmfile, array_length)
    return data
        
      
class DM4File:
    @property
    def endian_str(self):
        return self._endian_str 
    
    @property
    def hfile(self):
        return self._hfile
        
    def __init__(self, filedata):
        '''
        :param file filedata: file handle to dm4 file
        '''
        self._hfile = filedata
        self.header = read_header_dm4(self.hfile)
        self._endian_str = _get_struct_endian_str(self.header.little_endian)
            
        self.root_tag_dir_header = read_root_tag_dir_header_dm4(self.hfile, endian=self.endian_str)
        
    def close(self):
        self._hfile.close()
        self._hfile = None
             
    @classmethod
    def open(cls, filename):
        '''
        :param str filename: Name of DM4 file to open
        :rtype: DM4File
        :return: DM4File object
        '''
        
        hfile = open(filename, "rb")
        return DM4File(hfile)
    
    def read_tag_data(self, tag):
        '''Read the data associated with the passed tag'''
        return _read_tag_data(self.hfile, tag, self.endian_str)
     
    
    DM4TagDir = collections.namedtuple('DM4Dir', ('name', 'dm4_tag', 'named_subdirs', 'unnamed_subdirs','named_tags', 'unnamed_tags'))
        
    def read_directory(self, directory_tag=None):
        '''
        Read the directories and tags from a dm4 file.  The first step in working with a dm4 file.
        :return: A named collection containing information about the directory
        ''' 
        
        if directory_tag is None:
            directory_tag = self.root_tag_dir_header
            
        dir_obj = DM4File.DM4TagDir(directory_tag.name, directory_tag, {},[], {}, [])
        
        for iTag in range(0, directory_tag.num_tags):
            tag = read_tag_header_dm4(self.hfile, self.endian_str)
            if tag is None:
                break
            
            if tag_is_directory(tag):
                if tag.name is None:
                    dir_obj.unnamed_subdirs.append(self.read_directory(tag))
                else:
                    dir_obj.named_subdirs[tag.name] = self.read_directory(tag) 
            else:
                if tag.name is None:
                    dir_obj.unnamed_tags.append(tag)            
                else:
                    dir_obj.named_tags[tag.name] = tag
                
        return dir_obj
        
    
        
     
    


"""
The Pycroscopy package.

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    analysis
    io
    processing
    viz

"""
from . import analysis
from .analysis import *
from . import io
from .io import *
from . import processing
from .processing import *
from . import viz
from .viz import *

from .__version__ import version as __version__
from .__version__ import date as __date__

__all__ = ['processing', 'analysis', 'io', 'viz', '__date__', '__version__']
__all__ += io.__all__
__all__ += processing.__all__
__all__ += analysis.__all__
__all__ += viz.__all__



"""
Created on Feb 9, 2016

@author: Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals

import os

import numpy as np
from skimage.measure import block_reduce

from .df_utils.io_image import read_image, read_dm3
from .translator import Translator
from .utils import generate_dummy_main_parms, build_ind_val_dsets
from ..hdf_utils import getH5DsetRefs, calc_chunks, link_as_main
from ..io_hdf5 import ioHDF5
from ..microdata import MicroDataGroup, MicroDataset


class MovieTranslator(Translator):
    """
    Translate Pytchography data from a set of images to an HDF5 file
    """
    def __init__(self, *args, **kwargs):
        super(MovieTranslator, self).__init__(*args, **kwargs)

        self.rebin = False
        self.bin_factor = 1
        self.hdf = None
        self.binning_func = self.__no_bin
        self.bin_func = None
        self.image_ext = None

    def translate(self, h5_path, image_path, bin_factor=None, bin_func=np.mean, start_image=0, image_type='.tif'):
        """
        Basic method that adds Movie data to existing hdf5 file

        Parameters
        ----------------
        h5_path : str
            Absolute path to where the HDF5 file should be located
        image_path : str
            Absolute path to folder holding the image files
        bin_factor : array_like of uint, optional
            Downsampling factor for each dimension.  Default is None.
        bin_func : callable, optional
            Function which will be called to calculate the return value
            of each block.  Function must implement an axis parameter,
            i.e. numpy.mean.  Ignored if bin_factor is None.  Default is
            numpy.mean.
        start_image : int, optional
            Integer denoting which image in the file path should be considered the starting
            point.  Default is 0, start with the first image on the list.

        Returns
        ----------
        h5_main : h5py.Dataset
            HDF5 Dataset object that contains the flattened images

        """
        self.image_ext = image_type

        # Open the hdf5 file and delete any contents
        try:
            hdf = ioHDF5(h5_path)
            hdf.clear()
        except:
            raise
        self.hdf = hdf

        '''
        Get the list of all files with the provided extension and the number of files in the list
        '''
        if os.path.isfile(image_path):
            file_list, image_parms = read_dm3(image_path)
            usize = image_parms['SuperScan-Height']
            vsize = image_parms['SuperScan-Width']
            data_type = file_list.dtype.type
            num_images = file_list.shape[0]-start_image

        else:
            file_list = self._parse_file_path(image_path, image_type)

            # Set up the basic parameters associated with this set of images
            (usize, vsize), data_type, image_parms = self._getimagesize(os.path.join(image_path, file_list[0]))

            num_images = len(file_list) - start_image

        '''
        Check if a bin_factor is given.  Set up binning objects if it is.
        '''
        if bin_factor is not None:
            self.rebin = True
            if isinstance(bin_factor, int):
                self.bin_factor = (bin_factor, bin_factor)
            elif len(bin_factor) == 2:
                self.bin_factor = tuple(bin_factor)
            else:
                raise ValueError('Input parameter `bin_factor` must be a length 2 array_like or an integer.\n' +
                                 '{} was given.'.format(bin_factor))
            usize = int(usize / self.bin_factor[0])
            vsize = int(vsize / self.bin_factor[1])
            self.binning_func = block_reduce
            self.bin_func = bin_func
            data_type = np.float32

        h5_main, h5_mean_spec, h5_ronch = self._setupH5(usize, vsize, np.float32, num_images, image_parms)


        self._read_data(file_list[start_image:],
                        h5_main, h5_mean_spec, h5_ronch, image_path)

        return h5_main

    def _read_data(self, image_stack, h5_main, h5_mean_spec, h5_ronch, image_path):
        """
        Iterates over the images in `file_list`, reading each image and downsampling if
        reqeusted, and writes the flattened image to file.  Also builds the Mean_Ronchigram
        and the Spectroscopic_Mean datasets at the same time.

        Parameters
        ----------
        image_stack : list of str
            List of all files in `image_path` that will be read
        h5_main : h5py.Dataset
            Dataset which will hold the Ronchigrams
        h5_mean_spec : h5py.Dataset
            Dataset which will hold the Spectroscopic Mean
        h5_ronch : h5py.Dataset
            Dataset which will hold the Mean Ronchigram
        image_path : str
            Absolute file path to the directory which hold the images

        Returns
        -------
        None
        """

        mean_ronch = np.zeros(h5_ronch.shape, dtype=np.float32)

        num_files = len(image_stack)

        if os.path.isfile(image_path):
            self.__save_dm3_frames(image_stack, h5_main, h5_mean_spec, h5_ronch, mean_ronch, num_files)
        else:
            self.__read_image_files(image_stack, h5_main, h5_mean_spec, h5_ronch, image_path, mean_ronch, num_files)

    def __save_dm3_frames(self, image_stack, h5_main, h5_mean_spec, h5_ronch, mean_ronch, num_frames):
        """

        :param image_stack:
        :param h5_main:
        :param h5_mean_spec:
        :param h5_ronch:
        :param mean_ronch:
        :param num_frames:
        :return:
        """
        for iframe, thisframe in enumerate(image_stack):
            selected = (iframe + 1) % round(num_frames / 16) == 0
            if selected:
                print('Processing file...{}% - reading: {}'.format(round(100 * iframe / num_frames), iframe))
            image = self.binning_func(thisframe, self.bin_factor, self.bin_func).flatten()
            h5_main[:, iframe] = image

            h5_mean_spec[iframe] = np.mean(image)

            mean_ronch += image

            self.hdf.flush()

        h5_ronch[:] = mean_ronch / num_frames
        self.hdf.flush()

    def __read_image_files(self, image_stack, h5_main, h5_mean_spec, h5_ronch, image_path, mean_ronch, num_files):
        """
        Read each image from `file_list` and save it in `h5_main`.

        Parameters
        ----------
        image_stack:
        :param h5_main:
        :param h5_mean_spec:
        :param h5_ronch:
        :param image_path:
        :param mean_ronch:
        :param num_files:
        :return:
        """
        for ifile, thisfile in enumerate(image_stack):

            selected = (ifile + 1) % round(num_files / 16) == 0
            if selected:
                print('Processing file...{}% - reading: {}'.format(round(100 * ifile / num_files), thisfile))

            image = read_image(os.path.join(image_path, thisfile), greyscale=True)
            image = self.binning_func(image, self.bin_factor, self.bin_func)
            image = image.flatten()
            h5_main[:, ifile] = image

            h5_mean_spec[ifile] = np.mean(image)

            mean_ronch += image

            self.hdf.flush()
        h5_ronch[:] = mean_ronch / num_files
        self.hdf.flush()

    def downSampRoncVec(self, ronch_vec, binning_factor):
        """
        Downsample the image by taking the mean over nearby values

        Parameters
        ----------
        ronch_vec : ndarray
            Image data
        binning_factor : int
            factor to reduce the size of the image by

        Returns
        -------
        ronc_mat3_mean : ndarray
            Flattened downsampled image
        """
        ccd_pix = int(np.sqrt(ronch_vec.size))
        ronc_mat = ronch_vec.reshape(ccd_pix, ccd_pix)
        ronc_mat2 = ronc_mat.reshape(ccd_pix, ccd_pix / binning_factor, binning_factor)
        ronc_mat2_mean = ronc_mat2.mean(2)  # take the mean along the 3rd dimension
        ronc_mat3 = ronc_mat2_mean.reshape(ccd_pix / binning_factor, binning_factor, -1)
        ronc_mat3_mean = ronc_mat3.mean(1)

        return ronc_mat3_mean.reshape(-1)

    @staticmethod
    def _parse_file_path(path, ftype='all'):
        """
        Returns a list of all files in the directory given by path
        
        Parameters
        ---------------
        path : string / unicode
            absolute path to directory containing files
        ftype : this file types to return in file_list. (optional. Default is all) 
        
        Returns
        ----------
        file_list : list of strings
            names of all files in directory located at path
        numfiles : unsigned int
            number of files in file_list
        """    
        
        # Get all files in directory
        file_list = os.listdir(path)
        
        # If no file type specified, return full list
        if ftype == 'all':
            return file_list

        # Remove files of type other than the request ftype from the list
        new_file_list = []
        for this_thing in file_list:
            # Make sure it's really a file
            if not os.path.isfile(os.path.join(path, this_thing)):
                continue

            split = os.path.splitext(this_thing)
            ext = split[1]
            if ext == ftype:
                new_file_list.append(os.path.join(path, this_thing))

        return new_file_list

    @staticmethod
    def _getimagesize(image):
        """
        Returns the x and y size of the image in pixels
        
        Parameters
        ------------
        image : string / unicode
            absolute path to the image file
        
        Returns
        -----------
        (size, tmp.dtype) : Tuple 
        
        size : unsigned integer
            x and y dimenstions of image
        dtype : data type
            Datatype of the image
        """
        tmp, parms = read_image(image, get_parms=True)
        size = tmp.shape
        
        return size, tmp.dtype.type, parms

    def _setupH5(self, usize, vsize, data_type, num_images, main_parms):
        """
        Setup the HDF5 file in which to store the data including creating
        the Position and Spectroscopic datasets

        Parameters
        ----------
        usize : int
            Number of pixel columns in the images
        vsize : int
            Number of pixel rows in the images
        data_type : type
            Data type to save image as
        num_images : int
            Number of images in the movie
        main_parms : dict


        Returns
        -------
        h5_main : h5py.Dataset
            HDF5 Dataset that the images will be written into
        h5_mean_spec : h5py.Dataset
            HDF5 Dataset that the mean over all positions will be written
            into
        h5_ronch : h5py.Dataset
            HDF5 Dateset that the mean over all Spectroscopic steps will be
            written into
        """
        num_pixels = usize*vsize

        root_parms = generate_dummy_main_parms()
        root_parms['data_type'] = 'PtychographyData'

        main_parms['num_images'] = num_images
        main_parms['image_size_u'] = usize
        main_parms['image_size_v'] = vsize
        main_parms['num_pixels'] = num_pixels
        main_parms['translator'] = 'Movie'
    # Create the hdf5 data Group
        root_grp = MicroDataGroup('/')
        root_grp.attrs = root_parms
        meas_grp = MicroDataGroup('Measurement_000')
        meas_grp.attrs = main_parms
        chan_grp = MicroDataGroup('Channel_000')
    # Get the Position and Spectroscopic Datasets
    #     ds_spec_ind, ds_spec_vals = buildspectroscopicdatasets(usize, vsize, num_pixels)
        ds_spec_ind, ds_spec_vals = build_ind_val_dsets([num_images],
                                                        is_spectral=True,
                                                        labels=['Time'],
                                                        units=['s'])
        ds_pos_ind, ds_pos_val = build_ind_val_dsets([usize, vsize],
                                                     is_spectral=False,
                                                     labels=['X', 'Y'],
                                                     units=['pixel', 'pixel'])

        ds_chunking = calc_chunks([num_pixels, num_images],
                                  data_type(0).itemsize,
                                  unit_chunks=(num_pixels, 1))

    # Allocate space for Main_Data and Pixel averaged Data
        ds_main_data = MicroDataset('Raw_Data', data=[], maxshape=(num_pixels, num_images),
                                    chunking=ds_chunking, dtype=data_type, compression='gzip')
        ds_mean_ronch_data = MicroDataset('Mean_Ronchigram',
                                          data=np.zeros(num_pixels, dtype=np.float32),
                                          dtype=np.float32)
        ds_mean_spec_data = MicroDataset('Spectroscopic_Mean',
                                         data=np.zeros(num_images, dtype=np.float32),
                                         dtype=np.float32)
    # Add datasets as children of Measurement_000 data group
        chan_grp.addChildren([ds_main_data, ds_spec_ind, ds_spec_vals, ds_pos_ind,
                              ds_pos_val, ds_mean_ronch_data, ds_mean_spec_data])
        meas_grp.addChildren([chan_grp])

        root_grp.addChildren([meas_grp])
        # print('Writing following tree to this file:')
        # root_grp.showTree()

        h5_refs = self.hdf.writeData(root_grp)
        h5_main = getH5DsetRefs(['Raw_Data'], h5_refs)[0]
        h5_ronch = getH5DsetRefs(['Mean_Ronchigram'], h5_refs)[0]
        h5_mean_spec = getH5DsetRefs(['Spectroscopic_Mean'], h5_refs)[0]
        aux_ds_names = ['Position_Indices',
                        'Position_Values',
                        'Spectroscopic_Indices',
                        'Spectroscopic_Values']

        link_as_main(h5_main, *getH5DsetRefs(aux_ds_names, h5_refs))

        self.hdf.flush()
        
        return h5_main, h5_mean_spec, h5_ronch

    @staticmethod
    def __no_bin(image, *args, **kwargs):
        """
        Does absolutely nothing to the image.  Exists so that we can have
        a bin function to call whether we actually rebin the image or not.

        Parameters
        ----------
        image : ndarray
            Image
        args:
            Argument list
        kwargs:
            Keyword argument list

        Returns
        -------
        image : ndarray
            The input image
        """
        return image



"""
Pycroscopy's I/O module

Submodules
----------

.. autosummary::
    :toctree: _autosummary

    be_hdf_utils
    hdf_utils
    io_hdf5
    io_utils
    microdata
    translators

"""
from . import be_hdf_utils
from . import hdf_utils
from . import io_hdf5
from . import io_utils
from . import microdata
from . import translators
from .io_hdf5 import ioHDF5
from .io_utils import *
from .microdata import MicroDataset, MicroDataGroup
from .translators import *

__all__ = ['ioHDF5', 'MicroDataset', 'MicroDataGroup', 'be_hdf_utils', 'hdf_utils', 'io_utils', 'microdata']
__all__ += translators.__all__



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 05 07:55:56 2016

@author: Suhas Somnath, Chris Smith, Rama K. Vasudevan
"""

from __future__ import division, print_function, absolute_import, unicode_literals

from os import path
from warnings import warn

import h5py
import matplotlib.pyplot as plt
import numpy as np
import xlrd as xlreader

from ...be_hdf_utils import getActiveUDVSsteps, maxReadPixels
from ...hdf_utils import getAuxData, getDataSet, getH5DsetRefs, linkRefs, get_attr, create_spec_inds_from_vals
from ...io_hdf5 import ioHDF5
from ...io_utils import getAvailableMem, recommendCores
from ...microdata import MicroDataset, MicroDataGroup
from ....analysis.optimize import Optimize
from ....processing.proc_utils import buildHistogram
from ....viz.plot_utils import plot_1d_spectrum, plot_2d_spectrogram, plot_histgrams

nf32 = np.dtype({'names': ['super_band', 'inter_bin_band', 'sub_band'],
                 'formats': [np.float32, np.float32, np.float32]})


def parmsToDict(filepath, parms_to_remove=[]):
    """
    Translates the parameters in the text file into a dictionary. 
    Also indentifies whether this is a BEPS or BELine dataset.
    
    Parameters
    -----------
    filepath : String / Unicode
        Absolute path of the parameters text or spreadsheet file.
    parms_to_remove : List of string (Optional)
        keys that this function should attempt to remove from the dictionary
    
    Returns
    ----------
    isBEPS : Boolean
        whether this dataset is BEPS or BE Line.
    parm_dict : Dictionary
        experimental parameters
    """
    verbose = False

    lines = list()

    if filepath.lower().endswith('.txt'):
        file_handle = open(filepath, 'r')
        raw_lines = file_handle.readlines()
        file_handle.close()
        for line in raw_lines:
            line = line.rstrip()
            if len(line) > 0:
                lines.append(line.split(" : "))

    elif filepath.lower().endswith('.xls') or filepath.lower().endswith('.xlsx'):
        workbook = xlreader.open_workbook(filepath)
        worksheet = workbook.sheet_by_index(0)
        for row in range(worksheet.nrows):
            temp = list()
            for col in range(worksheet.ncols):
                try:
                    val = str(worksheet.cell(row, col).value).strip()
                    if len(val) > 0:
                        temp.append(val)
                except ValueError:
                    pass
                except:
                    raise
            lines.append(temp)
    else:
        warn('Parameter file not of expected format: text or spreadsheet')
        return None

    if verbose:
        print("Finished reading the file")

    is_beps = False
    parm_dict = dict()

    prefix = 'File_'
    for fields in lines:
        # Ignore the parameters describing the GUI choices
        if prefix == 'Multi_':
            continue

        # Check if the line is a group header or parameter/value pair
        if len(fields) == 2:
            # Get the current name/value pair, and clean up the name
            name = fields[0].strip().replace('# ', 'num_').replace('#', 'num_').replace(' ', '_')
            value = fields[1]

            # Rename specific parameters
            if name == '1_mode':
                name = 'mode'
            if name == 'IO_rate':
                name = 'IO_rate_[Hz]'
                value = int(value.split()[0])*1E6
            if name == 'AO_range':
                name = 'AO_range_[V]'
                value = ' '.join(value.split()[:-1])
            if name == 'AO_amplifier':
                value = value.split()[0]
            if name == 'cycle_time[s]':
                name = 'cycle_time_[s]'
            if name == 'FORC_V_high1[V]':
                name = 'FORC_V_high1_[V]'
            if name == 'FORC_V_high2[V]':
                name = 'FORC_V_high2_[V]'
            if name == 'FORC_V_low1[V]':
                name = 'FORC_V_low1_[V]'
            if name == 'FORC_V_low2[V]':
                name = 'FORC_V_low2_[V]'
            if name == 'amplitude[V]':
                name = 'amplitude_[V]'
            if name == 'offset[V]':
                name = 'offset_[V]'
            if name == 'set_pulst_amplitude[V]':
                name = 'set_pulst_amplitude_[V]'
            if name == 'set_pulst_duration[s]':
                name = 'set_pulst_duration_[s]'
            if name == 'step_edge_smoothing[s]':
                name = 'step_edge_smoothing_[s]'

            # Append the prefix to the name
            name = prefix+name.lstrip(prefix)

            # Write parameter to parm_dict
            try:
                num = float(value)
                parm_dict[name] = num
                if num == int(num):
                    parm_dict[name] = int(num)
            except ValueError:
                parm_dict[name] = value
            except:
                raise
        elif len(fields) == 1:
            # Change the parameter prefix to the new one from the group header
            prefix = fields[0].strip('<').strip('>')
            prefix = prefix.split()[0]+'_'

            # Check if there are VS Parameters.  Set isBEPS to true if so.
            is_beps = is_beps or prefix == 'VS_'

    if is_beps:
        useless_parms = \
            ['Multi_1_BE_response_spectra',
             'Multi_2_BE_amplitude_spectra',
             'Multi_3_BE_phase_spectra',
             'Multi_4_BE_response_spectrogram',
             'Multi_5_BE_amplitude_spectrogram',
             'Multi_6_BE_phase_spectrogram',
             'Multi_7_VS_response_loops',
             'Multi_8_VS_amplitude_loops',
             'Multi_9_VS_phase_loops',
             'Multi_10_topography',
             'Multi_11_mean_channel_2']
    else:
        useless_parms = \
            ['Multi_1_BE_response_spectra',
             'Multi_2_BE_amplitude_spectra',
             'Multi_3_BE_phase_spectra',
             'Multi_4_BE_response_spectrogram',
             'Multi_5_BE_amplitude_spectrogram',
             'Multi_6_BE_phase_spectrogram',
             'Multi_7_amplitude_map',
             'Multi_8_resonance_map',
             'Multi_9_Q_map',
             'Multi_10_phase_map',
             'Multi_11_topography',
             'Multi_12_AI_time_domain',
             'Multi_13_AI_Fourier_amplitude']

    if verbose:
        print("Finished parsing the text pairs. isBEPS = {}".format(is_beps))

    useless_parms.extend(parms_to_remove)

    # Now remove the list of useless parameters:
    for uparm in useless_parms:
        try:
            del parm_dict[uparm]
        except KeyError:
            #warn('Parameter to be deleted does not exist')
            pass
        except:
            raise
    del uparm, useless_parms

    if verbose:
        print("Finished removing useless parameters")

    if is_beps:
        # fix the DC type in the parms:
        if parm_dict['VS_measure_in_field_loops'] == 'out-of-field only':
            parm_dict['VS_measure_in_field_loops'] = 'out-of-field'
        elif parm_dict['VS_measure_in_field_loops'] == 'in-field only':
            parm_dict['VS_measure_in_field_loops'] = 'in-field'

    return is_beps, parm_dict
    
###############################################################################


def requires_conjugate(chosen_spectra, default_q=10):
    """
    Determines whether or not the conjugate of the data needs to be taken based on the quality factor

    Parameters
    ----------
    chosen_spectra : 2D complex numpy array
        N random spectra arranged as [instance, frequency]
    default_q : unsigned int, Optional
        Default value of Q factor that the SHO guess function results in for poor guesses

    Returns
    -------
    do_conjugate : Boolean
        Whether or not to take the conjugate of the data
    """
    # Do the SHO Guess for each of these
    opt = Optimize(data=chosen_spectra)

    fitguess_results = opt.computeGuess(strategy='complex_gaussian',
                                        processors=recommendCores(chosen_spectra.shape[0]),
                                        options={'frequencies': np.arange(chosen_spectra.shape[1])})

    q_results = np.array(fitguess_results)[:, 2]
    good_q = q_results[np.where(q_results != default_q)]

    if np.mean(good_q) < 0:
        return True
    else:
        return False


###############################################################################

def getSpectroscopicParmLabel(expt_type):
    """
    Returns the label for the spectroscopic parameter in the plot group. 
    
    Parameters
    ----------
    expt_type : str
        Type of the experiment - found in the parms.txt file
    
    Returns
    -------
    str
        label for the spectroscopic parameter axis in the plot
    """
    
    if expt_type in ['DC modulation mode', 'current mode']:
        return 'DC Bias'
    elif expt_type == 'AC modulation mode with time reversal':
        return 'AC amplitude'
    return 'User Defined'

###############################################################################

def normalizeBEresponse(spectrogram_mat, FFT_BE_wave, harmonic):
    """
    This function normalizes the BE waveform to correct the phase by diving by the excitation
    
    Parameters
    ------------
    spectrogram_mat : 2D complex numpy array
        BE response arranged as [bins, steps]
    FFT_BE_wave : 1D complex numpy array
        FFT of the BE waveform at the appropriate bins. Number of bins must match with spectrogram_mat
    harmonic : unsigned int
        nth harmonic of the excitation waveform
        
    Returns
    ----------
    spectrogram_mat : 2D complex numpy array
        Normalized BE response spectrogram

    """  
    BE_wave = np.fft.ifftshift(np.fft.ifft(FFT_BE_wave))
    scaling_factor = 1
      
    if harmonic == 2:
        scaling_factor = np.fft.fftshift(np.fft.fft(BE_wave**2))/(2*np.exp(1j*3*np.pi*0.5))
    elif harmonic == 3:
        scaling_factor = np.fft.fftshift(np.fft.fft(BE_wave**3))/(4*np.exp(1j*np.pi))
    elif harmonic >= 4:
        print("Warning these high harmonics are not supported in translator.")
 
    # Generate transfer functions
    F_AO_spectrogram = np.transpose(np.tile(FFT_BE_wave/scaling_factor, [spectrogram_mat.shape[1], 1]))
    # Divide by transfer function
    spectrogram_mat = spectrogram_mat/(F_AO_spectrogram)
  
    return spectrogram_mat

    
def generatePlotGroups(h5_main, hdf, mean_resp, folder_path, basename, max_resp=[], min_resp=[], 
                       max_mem_mb=1024, spec_label='None', ignore_plot_groups=[], 
                       show_plots=True, save_plots=True, do_histogram=False,
                       debug=False):
    """
    Generates the spatially averaged datasets for the given raw dataset. 
    The averaged datasets are necessary for quick visualization of the quality of data. 
    
    Parameters
    ----------
    h5_main : H5 reference
        to the main dataset
    hdf : active ioHDF instance 
        for writing to same H5 file
    mean_resp : 1D numpy array
        spatially averaged amplitude
    folder_path : String
        Absolute path of the data folder
    basename : String
        base name of the dataset
    max_resp : 1D numpy array
        Maximum amplitude for all pixels
    min_resp : 1D numpy array
        Minimum amplitude for all pixels
    max_mem_mb : Unisigned integer
        Maximum memory that can be used for generating histograms
    spec_label : String
        Parameter that is varying
    ignore_plot_groups : (optional) List of strings
        Names of the plot groups (UDVS columns) that should be ignored
    show_plots : (optional) Boolean
        Whether or not to show plots
    save_plots : (optional) Boolean
        Whether or not to save generated plots
    do_histogram : Boolean (Optional. Default = False)
        Whether or not to generate hisograms. 
        Caution - Histograms can take a fair amount of time to compute.
    """

    grp = h5_main.parent
    h5_freq = grp['Bin_Frequencies']    
    UDVS = grp['UDVS']
    spec_inds = grp['Spectroscopic_Indices']
    UDVS_inds = grp['UDVS_Indices']
    spec_vals = grp['Spectroscopic_Values']
                    
#     std_cols = ['wave_type','Frequency','DC_Offset','wave_mod','AC_Amplitude','dc_offset','ac_amplitude']
    
    col_names = UDVS.attrs['labels']
    
    if len(col_names) <= 5:
        """
        No plot groups are defined in the UDVS table.
        All plot group datasets will be written to the 
        Channel group
        """
        return
    
    # Removing the standard columns
    col_names = get_attr(UDVS, 'labels')[5:]

#     col_names = [col for col in col_names if col not in std_cols + ignore_plot_groups]
    
    freq_inds = spec_inds[spec_inds.attrs['Frequency']].flatten()
    
    for col_name in col_names:
        ref = UDVS.attrs[col_name]
#         Make sure we're actually dealing with a reference of some type
        if not isinstance(ref, h5py.RegionReference):
            continue
        # 4. Access that column of the data through region reference
        steps = np.where(np.isfinite(UDVS[ref]))[0]
        step_inds = np.array([np.where(UDVS_inds.value == step)[0] for step in steps]).flatten()
        """selected_UDVS_steps = UDVS[ref]
        selected_UDVS_steps = selected_UDVS_steps[np.isfinite(selected_UDVS_steps)]"""

        (step_averaged_vec, mean_spec) = reshape_mean_data(spec_inds, step_inds, mean_resp)
            
        """ 
        Need to account for cases with multiple excitation waveforms
        This will affect the frequency indices / values
        We are assuming that there is only one excitation waveform per plot group
        """
        freq_slice = np.unique(freq_inds[step_inds])
        freq_vec = h5_freq.value[freq_slice]
        
        num_bins = len(freq_slice)  # int(len(freq_inds)/len(UDVS[ref]))
        pg_data = np.repeat(UDVS[ref], num_bins)
            
        ds_mean_spec = MicroDataset('Mean_Spectrogram', mean_spec, dtype=np.complex64)
        ds_step_avg = MicroDataset('Step_Averaged_Response', step_averaged_vec, dtype=np.complex64)
        # cannot assume that this is DC offset, could be AC amplitude....
        ds_spec_parm = MicroDataset('Spectroscopic_Parameter', np.squeeze(pg_data[step_inds]))
        ds_spec_parm.attrs = {'name': spec_label}
        ds_freq = MicroDataset('Bin_Frequencies', freq_vec)
        
        plot_grp = MicroDataGroup('{:s}'.format('Spatially_Averaged_Plot_Group_'), grp.name[1:])
        plot_grp.attrs['Name'] = col_name
        plot_grp.addChildren([ds_mean_spec, ds_step_avg, ds_spec_parm, ds_freq])
        
        h5_plt_grp_refs = hdf.writeData(plot_grp, print_log=debug)
        
        h5_mean_spec = getH5DsetRefs(['Mean_Spectrogram'], h5_plt_grp_refs)[0]
        h5_step_avg = getH5DsetRefs(['Step_Averaged_Response'], h5_plt_grp_refs)[0]
        h5_spec_parm = getH5DsetRefs(['Spectroscopic_Parameter'], h5_plt_grp_refs)[0]
        h5_freq_vec = getH5DsetRefs(['Bin_Frequencies'], h5_plt_grp_refs)[0]
        
        # Linking the datasets with the frequency and the spectroscopic variable:
        linkRefs(h5_mean_spec, [h5_spec_parm, h5_freq_vec])
        linkRefs(h5_step_avg, [h5_freq_vec])
        
        """
        Create Region Reference for the plot group in the Raw_Data, Spectroscopic_Indices 
        and Spectroscopic_Values Datasets
        """
        raw_ref = h5_main.regionref[:, step_inds]
        spec_inds_ref = spec_inds.regionref[:, step_inds]
        spec_vals_ref = spec_vals.regionref[:, step_inds]
        
        ref_name = col_name.replace(' ', '_').replace('-', '_')+'_Plot_Group'
        h5_main.attrs[ref_name] = raw_ref
        spec_inds.attrs[ref_name] = spec_inds_ref
        spec_vals.attrs[ref_name] = spec_vals_ref
        
        hdf.flush()
        
        if do_histogram:
            """
            Build the histograms for the current plot group
            """
            hist = BEHistogram()
            hist_mat, hist_labels, hist_indices, hist_indices_labels = \
                hist.buildPlotGroupHist(h5_main, step_inds, max_response=max_resp,
                                        min_response=min_resp, max_mem_mb=max_mem_mb, debug=debug)
            ds_hist = MicroDataset('Histograms', hist_mat, dtype=np.int32,
                                   chunking=(1, hist_mat.shape[1]),compression='gzip')
            hist_slice_dict = dict()
            for hist_ind, hist_dim in enumerate(hist_labels):
                hist_slice_dict[hist_dim] = (slice(hist_ind, hist_ind+1), slice(None))
            ds_hist.attrs['labels'] = hist_slice_dict
            ds_hist.attrs['units'] = ['V', '', 'V', 'V']
            ds_hist_indices = MicroDataset('Indices', hist_indices, dtype=np.uint)
            ds_hist_values = MicroDataset('Values', hist_indices, dtype=np.float32)
            hist_ind_dict = dict()
            for hist_ind_ind, hist_ind_dim in enumerate(hist_indices_labels):
                hist_ind_dict[hist_ind_dim] = (slice(hist_ind_ind, hist_ind_ind+1), slice(None))
            ds_hist_indices.attrs['labels'] = hist_ind_dict
            ds_hist_values.attrs['labels'] = hist_ind_dict

            hist_grp = MicroDataGroup('Histogram', h5_mean_spec.parent.name[1:])

            hist_grp.addChildren([ds_hist, ds_hist_indices, ds_hist_values])
            
            h5_hist_grp_refs = hdf.writeData(hist_grp)
            
            h5_hist = getH5DsetRefs(['Histograms'], h5_hist_grp_refs)[0]
            h5_hist_inds = getH5DsetRefs(['Indices'], h5_hist_grp_refs)[0]
            h5_hist_vals = getH5DsetRefs(['Values'], h5_hist_grp_refs)[0]
            
            linkRefs(h5_hist, getH5DsetRefs(['Indices', 'Values'], h5_hist_grp_refs))
            
            h5_hist.attrs['Spectroscopic_Indices'] = h5_hist_inds.ref
            h5_hist.attrs['Spectroscopic_Values'] = h5_hist_vals.ref
                
        else:
            """
            Write the min and max response vectors so that histograms can be generated later.
            """
            ds_max_resp = MicroDataset('Max_Response', max_resp)
            ds_min_resp = MicroDataset('Min_Response', min_resp)
            plot_grp.addChildren([ds_max_resp, ds_min_resp])
        
        if save_plots or show_plots:
            fig_title = '_'.join(grp.name[1:].split('/')+[col_name])
            path_1d = None
            path_2d = None
            path_hist = None
            if save_plots:
                path_1d = path.join(folder_path, basename + '_Step_Avg_' + fig_title + '.png')
                path_2d = path.join(folder_path, basename + '_Mean_Spec_' + fig_title + '.png')
                path_hist = path.join(folder_path, basename + '_Histograms_' + fig_title + '.png')
            plot_1d_spectrum(step_averaged_vec, freq_vec, fig_title, figure_path=path_1d)
            plot_2d_spectrogram(mean_spec, freq_vec, fig_title, figure_path=path_2d)
            if do_histogram:
                plot_histgrams(hist_mat, hist_indices, grp.name, figure_path=path_hist)
            
            if show_plots:
                plt.show()
            plt.close('all')
        # print('Generated spatially average data for group: %s' %(col_name))
    print('Completed generating spatially averaged plot groups')

###############################################################################


def reshape_mean_data(spec_inds, step_inds, mean_resp):
    """
    Takes in the mean data vector and rearranges that data according to 
    plot group as [step number,bins]
    
    Parameters
    -----------
    spec_inds : 2D numpy array
        UDVS_Indices as a 2D mat
    step_inds : 1D numpy array or list
        UDVS step indices corresponding to this plot group
    mean_resp : 1D numpy complex array
        position averaged BE data
            
    Returns
    ----------
    step_averaged_vec : 1D complex numpy array
        Mean (position averaged) spectrogram averaged over the UDVS steps as well
    mean_spectrogram : 2D complex numpy array
        Position averaged data arranged as [step number,bins]
    """
    num_bins = len(np.unique(spec_inds[0, step_inds]))
    # Stephen says that we can assume that the number of bins will NOT change in a plot group
    mean_spectrogram = mean_resp[step_inds].reshape(-1, num_bins)
        
    step_averaged_vec = np.mean(mean_spectrogram, axis=0)
    return step_averaged_vec, mean_spectrogram
    
###############################################################################


def visualize_plot_groups(h5_filepath):
    """
    Visualizes the plot groups present in the provided BE data file
    
    Parameters
    ----------
    h5_filepath : String / Uniciode
        Absolute path of the h5 file
    """
    with h5py.File(h5_filepath, mode='r') as h5f:
        expt_type = h5f.attrs.get('data_type')
        if expt_type not in ['BEPSData', 'BELineData']:
            warn('Invalid data format')
            return         
        for grp_name in h5f.keys():            
            grp = h5f[grp_name]['Channel_000']            
            for plt_grp_name in grp.keys():                
                if plt_grp_name.startswith('Spatially_Averaged_Plot_Group_'): 
                    plt_grp = grp[plt_grp_name]
                    if expt_type == 'BEPSData':
                        spect_data = plt_grp['Mean_Spectrogram'].value
                        plot_2d_spectrogram(spect_data, plt_grp['Bin_Frequencies'].value, plt_grp.attrs['Name'])
                    step_avg_data = plt_grp['Step_Averaged_Response']
                    plot_1d_spectrum(step_avg_data, plt_grp['Bin_Frequencies'].value, plt_grp.attrs['Name'])
                    try:
                        hist_data = plt_grp['Histograms']
                        hist_bins = plt_grp['Histograms_Indicies']
                        plot_histgrams(hist_data, hist_bins, plt_grp.attrs['Name'])
                    except:
                        pass
                    
    plt.show()
    plt.close('all')
    
###############################################################################


def trimUDVS(udvs_mat, udvs_labs, udvs_units, target_col_names):
    """
    Removes unused (typically default) plot groups
    
    Parameters
    ----------
    udvs_mat : 2D numpy array
        UDVS table arranged as [steps, col]
    udvs_labs : list of strings
        Column names of the UDVS table
    udvs_units : list of strings
        Units for the columns of the UDVS table
    target_col_names : list of strings
        Column names that need to be removed
        
    Returns
    -----------
    udvs_mat : 2D numpy array
        Truncated UDVS table
    udvs_labs : list of strings 
        Truncated list of UDVS column names
    udvs_units : list of strings
        Truncated list of UDVS column units
    """
    
    if len(target_col_names) == 0:
        return udvs_mat, udvs_labs, udvs_units
        
    if len(udvs_labs) != udvs_mat.shape[1]:
        warn('Error: Incompatible UDVS matrix and labels. Not truncating!')
        return udvs_mat, udvs_labs, udvs_units
    
    # First figure out the column indices
    col_inds = []
    found_cols = []
    for target in target_col_names:
        for ind, udvs_col in enumerate(udvs_labs):
            if udvs_col == target:
                col_inds.append(ind)
                found_cols.append(udvs_col)
                break
    col_inds = np.sort(np.unique(col_inds))
    
    # Now remove from the labels and the matrix
    udvs_mat = np.delete(udvs_mat, col_inds, axis=1)
    # col_inds.sort(reverse=True)
    [udvs_units.pop(ind) for ind in range(len(col_inds), 0, -1)]
    udvs_labs = [col for col in udvs_labs if col not in found_cols]
    
    return udvs_mat, udvs_labs, udvs_units

###############################################################################

def createSpecVals(udvs_mat, spec_inds, bin_freqs, bin_wfm_type, parm_dict, 
                   udvs_labs, udvs_units):
    """
    This function will determine the proper Spectroscopic Value array for the 
    dataset
    
    Chris Smith -- csmith55@utk.edu
    
    Parameters
    ----------
    udvs_mat : numpy array 
        UDVS table from dataset
    spec_inds : numpy array 
        Spectroscopic Indices table from dataset
    bin_freqs : numpy array 
        Bin frequencies
    bin_wfm_type : numpy array 
        waveform type for each frequency index
    parm_dict : dictionary
        parameters for dataset
    udvs_labs : list of strings
        labels for the columns of the UDVS matrix
    udvs_units : list of strings
        units for the columns of the UDVS matrix
    
    Returns
    -----------
    ds_spec_val_mat : numpy array 
        Spectroscopic Values table
    ds_spec_val_labs : list of strings
        names of the columns of the Spectroscopic Values table
    ds_spec_val_units : list of strings
        units of the columns of the Spectroscopic Values table
    ds_spec_val_labs_names : list of Strings
        labels with the names of their parameters
    """
    def __FindSpecValIndices(udvs_mat, spec_inds, usr_defined=False):
        """
        This function finds the Spectroscopic Values associated with the dataset that
        have more than one unique value
        
        Parameters
        ----------
        udvs_mat : numpy array containing the UDVS table
        spec_inds : numpy array contain Spectroscopic indices table
    
        Returns
        -------
        iSpec_var : integer array holding column indices in UDVS that change
        ds_spec_val_mat : array holding all spectral values for columns in iSpec_var
        """
#         Copy even step values of DC_offset into odd steps 
        UDVS = np.copy(udvs_mat)

        if not usr_defined:
            DC = UDVS[:, 1]
            for step in range(0, DC.size, 2):
                DC[step+1] = DC[step]
            UDVS[:, 1] = DC
        
        """
        icheck is an array containing all UDVS steps which should be checked.
        """
        icheck = np.unique(spec_inds[1])
        """
        Keep only the UDVS values for steps which we care about and the 
        first 5 columns
        """
        UDVS = UDVS[icheck, :5]
#         UDVS = np.array([UDVS[i] for i in icheck])
        
        """
        Transpose UDVS for ease of looping later on and store the number of steps
        as num_cols
        """
        num_cols = np.size(UDVS, 1)
        """
        Initialize the iSpec_var as an empty array.  It will store the index of the 
        UDVS label for any column which has more than one unique value
        """    
        iSpec_var = []
        
        """
        Loop over all columns in udvs_mat
        """
        for i in range(1, num_cols):
            """
            Find all unique values in the current column
            """
            toosmall = np.where(abs(UDVS[:, i]) < 1E-5)[0]
            UDVS[toosmall, i] = 0
            uvals = np.unique(UDVS[:, i])
            """
            np.unique considers all NaNs to be unique values
            These two lines find the indices of all NaNs in the unique value array 
            and removes all but the first
            """
            nanvals = np.where(np.isnan(uvals))[0]
            uvals = np.delete(uvals, nanvals[1:])
            """
            Check if more that one unique value
            Append column number to iSpec_var if true
            """
            if (uvals.size > 1): 
                iSpec_var = np.append(iSpec_var, int(i))
        
        iSpec_var = np.asarray(iSpec_var, np.int)
        ds_spec_val_mat = UDVS[:, iSpec_var]
        
        return iSpec_var, ds_spec_val_mat
                    
    def __BEPSVals(udvs_mat, spec_inds, bin_freqs, bin_wfm_type, parm_dict, udvs_labs, udvs_units):
        """
        Returns the Spectroscopic Value array for a BEPS dataset
    
        Parameters
        ----------
        udvs_mat : hdf5 dataset reference to UDVS dataset
        spec_inds : numpy array containing Spectroscopic indices table 
        bin_freqs : 1D numpy array of frequencies
        bin_wfm_type : numpy array containing the waveform type for each frequency index
        parm_dict : parameter dictinary for dataset
        udvs_labs : list of labels for the columns of the UDVS matrix
        udvs_units : list of units for the columns of the UDVS matrix
        
        Returns
        -------
        ds_spec_val_mat : list holding final Spectroscopic Value table 
        """
        
        """
        Check the mode to determine if DC, AC, or something else
        """
        mode = parm_dict['VS_mode']
        
        if mode == 'DC modulation mode' or mode == 'current mode':
            """ 
            First we call the FindSpecVals function to get the columns in UDVS of  
            interest and return a first pass on the spectral value array 
            """ 
            iSpecVals, inSpecVals = __FindSpecValIndices(udvs_mat, spec_inds)   
            
            return __BEPSDC(udvs_mat,inSpecVals, bin_freqs, bin_wfm_type, parm_dict)

        elif mode == 'AC modulation mode with time reversal':
            """ 
            First we call the FindSpecVals function to get the columns in UDVS of  
            interest and return a first pass on the spectral value array 
            """ 
            iSpecVals, inSpecVals = __FindSpecValIndices(udvs_mat, spec_inds)   
                       
            return __BEPSAC(udvs_mat,inSpecVals, bin_freqs, bin_wfm_type, parm_dict)
        else:
            """ 
            First we call the FindSpecVals function to get the columns in UDVS of  
            interest and return a first pass on the spectral value array 
            """ 
            iSpecVals, inSpecVals = __FindSpecValIndices(udvs_mat, spec_inds, usr_defined=True)   
                                   
            return __BEPSgen(udvs_mat, inSpecVals, bin_freqs, bin_wfm_type,
                             parm_dict, udvs_labs, iSpecVals, udvs_units)

    def __BEPSDC(udvs_mat,inSpecVals, bin_freqs, bin_wfm_type, parm_dict):
        """
        Calculates Spectroscopic Values for BEPS data in DC modulation mode
        
        Parameters
        ----------
        udvs_mat : hdf5 dataset reference to UDVS dataset
        inSpecVals : list holding initial guess at spectral values 
        bin_freqs : 1D numpy array of frequencies
        bin_wfm_type : numpy array containing the waveform type for each frequency index
        parm_dict : parameter dictinary for dataset
                        
        Returns
        -------
        ds_spec_val_mat : list holding final Spectroscopic Value table 
        SpecValsLabels : list holding labels of column names of ds_spec_val_mat
        """
        hascycles = False
        hasFORCS = False
        
        # print('in shape:',np.shape(inSpecVals))
        """
        All DC datasets will need Spectroscopic Value fields for Bin, DC, and Field
        
        Bin     - Bin number
        DC      - dc offset from UDVS
        Field   - 0 if in-field, 1 if out-of-field 
        """
        nrow = 2
        ds_spec_val_labs = ['Frequency','DC_Offset']
        ds_spec_val_units = ['Hz', 'V']
        
        """
        Get the number of and steps in the current dataset
        """
        numsteps = udvs_mat.shape[0]
        
        """
        Get the wave form for each step from udvs_mat
        """
        wave_form = udvs_mat[:,3]        
        
        """
        Define list of attribute names needed from the group metadata
        """
        field_type = parm_dict['VS_measure_in_field_loops']
        numcycles = parm_dict['VS_number_of_cycles']
        numFORCs = parm_dict['FORC_num_of_FORC_cycles']
        numcyclesteps = parm_dict['VS_steps_per_full_cycle']
        cycle_fraction = parm_dict['VS_cycle_fraction']
        
        if field_type == 'in and out-of-field':
            ds_spec_val_labs.append('Field')
            ds_spec_val_units.append('')
            nrow += 1
        
        frac = {'full':1.0,'1/2':0.5,'1/4':0.25,'3/4':0.75}
        
        numcyclesteps = frac[cycle_fraction]*numcyclesteps
        
        """
        Check the number of cycles and FORCs
        Add to Spectroscopic Values Labels as needed 
        """
        if numcycles > 1:
            hascycles = True
            nrow += 1
            ds_spec_val_labs.append('Cycle')
            ds_spec_val_units.append('')
            
        if numFORCs > 1:
            hasFORCS = True
            """
        It's possible to have 1 cycle with multiple FORCs so force cycle tracking 
        if FROCs exist
            """
            nrow += 1
            ds_spec_val_labs.append('FORC')
            ds_spec_val_units.append('')
            numFORCsteps = numcycles*numcyclesteps*2
        
        """
        Check the field type
        For in-field and out-of-field we know all values of field ahead of time
        For in and out-of-field we must check at each step
        If something else is in field_type, we default to in and out-of-field and print message
        """
        if field_type == 'out-of-field':
            field = 1
            numsteps = int(numsteps/2)
            numcyclesteps = int(numcyclesteps/2)
            swapfield = [1,1]
            field_names = ['out-of-field']
        elif field_type == 'in-field':
            field = 0
            numsteps = int(numsteps/2)
            numcyclesteps = int(numcyclesteps/2)
            swapfield = [0,0]
            field_names = ['in-field']
        elif field_type == 'in and out-of-field':
            field = 0
            swapfield = [1,0]
            field_names = ['out-of-field','in-field']
        else:
            warn('{} is not a known field type'.format(field_type))
            field = 0
            swapfield = [1,0]
            
        """
        Initialize ds_spec_val_mat so that we can append to it in loop
        """
        ds_spec_val_mat = np.empty([nrow,1])
    
        """
        Main loop over all steps
        """
        FORC=-1
        cycle=-1
        for step in range(numsteps):
            """
            Calculate the cycle number if needed
            """
            if hasFORCS:
                FORC = np.floor(step/numFORCsteps)
                stepinFORC = step-FORC*numFORCsteps
                cycle = np.floor(stepinFORC/numcyclesteps/2)
            elif hascycles:
                cycle = np.floor(step/numcyclesteps/2)
            
            """
            Change field if needed
            """
            field = swapfield[field]
            """
            Get bins for current step based on waveform
            """
            this_wave = np.where(bin_wfm_type==wave_form[step])[0]
            
            """
            Loop over bins
            """
            for thisbin in this_wave:
                colVal = np.array([[bin_freqs[thisbin]],[inSpecVals[step][0]]])
                
                if field_type == 'in and out-of-field':
                    colVal = np.append(colVal, [[field]], axis=0)
                """
                Add entries to cycle and/or FORC as needed
                """
                if hascycles:
                    colVal = np.append(colVal, [[cycle]], axis=0)
                if hasFORCS:
                    colVal = np.append(colVal, [[FORC]], axis=0)
                
                ds_spec_val_mat = np.append(ds_spec_val_mat, colVal, axis=1)
    
        return ds_spec_val_mat[:,1:], ds_spec_val_labs, ds_spec_val_units, [['Field',field_names]]
        
    def __BEPSAC(udvs_mat, inSpecVals, bin_freqs, bin_wfm_type, parm_dict):
        """
        Calculates Spectroscopic Values for BEPS data in AC modulation mode with time 
                reversal
        
        Parameters
        ----------
        udvs_mat : hdf5 dataset reference to UDVS dataset
        inSpecVals : list holding initial guess at spectral values 
        bin_freqs : 1D numpy array of frequencies
        bin_wfm_type : numpy array containing the waveform type for each frequency index
        parm_dict : parameter dictinary for dataset            
            
        Returns
        -------
        ds_spec_val_mat : list holding final Spectroscopic Value table 
        SpecValsLabels : list holding labels of column names of ds_spec_val_mat
        """
        
        hascycles = False
        hasFORCS = False
    
        """
        All AC datasets will need Spectroscopic Value fields for Bin, AC, and Direction
        
        Bin     - Bin number
        AC      - AC amplitude from UDVS
        forrev   - 1 if forward, -1 if reverse 
        """
        nrow = 3
        ds_spec_val_labs = ['Frequency','AC_Amplitude','Direction']
        ds_spec_val_units = ['Hz', 'A', '']
        
        """
        Get the number of bins and steps in the current dataset
        """
        numsteps = np.shape(udvs_mat)[0]

        """
        Get the wave form for each step from udvs_mat
        """
        wave_form = udvs_mat[:,3]        
        
        """
        Define list of attribute names needed from the group metadata
        """
        numcycles = parm_dict['VS_number_of_cycles']
        numFORCs = parm_dict['FORC_num_of_FORC_cycles']
        numcyclesteps = parm_dict['VS_steps_per_full_cycle']
        cycle_fraction = parm_dict['VS_cycle_fraction']
        
        frac = {'full':1.0,'1/2':0.5,'1/4':0.25,'3/4':0.75}
        
        numcyclesteps = frac[cycle_fraction]*numcyclesteps
        
        """
        Check the number of cycles and FORCs
        Add to Spectroscopic Values Labels as needed 
        """
        if numcycles > 1:
            hascycles = True
            nrow += 1
            ds_spec_val_labs.append('Cycle')
            ds_spec_val_units.append('')
            
        if numFORCs > 1:
            hasFORCS = True
            nrow += 1
            ds_spec_val_labs.append('FORC')
            ds_spec_val_units.append('')
            numFORCsteps = numcycles*numcyclesteps

        """
        Initialize ds_spec_val_mat so that we can append to it in loop
        """
        ds_spec_val_mat = np.empty([nrow,1])

        """
        Main loop over all steps
        """
        FORC=-1
        cycle=-1
        for step in range(numsteps):
            """
            Calculate the cycle number if needed
            """
            if hasFORCS:
                FORC = np.floor(step/numFORCsteps)
                stepinFORC = step-FORC*numFORCsteps
                cycle = np.floor(stepinFORC/numcyclesteps)
            elif hascycles:
                cycle = np.floor(step/numcyclesteps)
            
            """
            Check the wave_mod
            """
            wmod = inSpecVals[step][1]
            forrev = np.sign(wmod)

            """
            Get bins for current step based on waveform
            """
            this_wave = np.where(bin_wfm_type==wave_form[step])[0]
            
            """
            Loop over bins
            """
            for thisbin in this_wave:
                colVal = np.array([[bin_freqs[thisbin]],[inSpecVals[step][0]],[forrev]])
                """
                Add entries to cycle and/or FORC as needed
                """
                if hascycles:
                    colVal = np.append(colVal, [[cycle]], axis=0)
                if hasFORCS:
                    colVal = np.append(colVal, [[FORC]], axis=0)
                
                ds_spec_val_mat = np.append(ds_spec_val_mat, colVal, axis=1)
    
        return ds_spec_val_mat[:,1:], ds_spec_val_labs, ds_spec_val_units, [['Direction',['reverse','forward']]]
    
    
    def __BEPSgen(udvs_mat,inSpecVals, bin_freqs, bin_wfm_type, parm_dict, udvs_labs, iSpecVals, udvs_units):
        """
        Calculates Spectroscopic Values for BEPS data in generic mode
        
        Parameters
        ----------
        udvs_mat : hdf5 dataset reference to UDVS dataset
        inSpecVals : list holding initial guess at spectral values 
        bin_freqs : 1D numpy array of frequencies
        bin_wfm_type : numpy array containing the waveform type for each frequency index
        parm_dict : parameter dictinary for dataset            
        udvs_labs : list of labels for the columns of the UDVS matrix
            
        Returns
        -------
        ds_spec_val_mat -- list holding final Spectroscopic Value table 
        SpecValsLabels -- list holding labels of column names of ds_spec_val_mat
        """
        

        """
        Get the number of bins and steps in the current dataset
        """
        numsteps = udvs_mat.shape[0]
        
        """
        Get the wave form for each step from udvs_mat
        """
        wave_form = udvs_mat[:,3]        
        
        """
        All datasets will need Spectroscopic Value fields for Bin,
        everything else must be defined
        
        Bin     - Bin number
        """
        ds_spec_val_labs = ['Frequency']
        ds_spec_val_units = ['Hz']
        
        ds_spec_val_labs.extend(udvs_labs[(iSpecVals[:])])
        ds_spec_val_units.extend([udvs_units[i] for i in iSpecVals])
        nrow = len(ds_spec_val_labs)
        
        """
        Initialize ds_spec_val_mat so that we can append to it in loop
        """
        ds_spec_val_mat = np.empty([nrow,1])
    
        """
        Main loop over all steps
        """
        for step in range(numsteps):
            """
            Get the wave form for each step from udvs_mat
            """
            this_wave = np.where(bin_wfm_type==wave_form[step])[0]
            
            for thisbin in this_wave:
                colVal = np.array([[bin_freqs[thisbin]]])
                colVal = np.append(colVal, [[row] for row in inSpecVals[step,:]], axis=0)
                
                ds_spec_val_mat = np.append(ds_spec_val_mat, colVal, axis=1)
    
        return ds_spec_val_mat[:,1:], ds_spec_val_labs, ds_spec_val_units, []

    """
********************************************************************************************
    END OF INTERNAL FUNCTION LIST
    """
    dtype = parm_dict['data_type']
    if dtype == 'BELineData':
        ds_spec_val_mat = bin_freqs.reshape([1,-1])
        ds_spec_inds_mat = np.zeros_like(ds_spec_val_mat, dtype = np.int32)
        ds_spec_val_labs = ['Frequency']
        ds_spec_val_units = ['Hz']
        spec_vals_labs_names = []
        ds_spec_inds_mat[0,:] = np.arange(ds_spec_val_mat.shape[1])
        
    elif dtype == 'BEPSData':
        # Call __BEPSVals to finish the refining of the Spectroscopic Value array        
        ds_spec_val_mat, ds_spec_val_labs, ds_spec_val_units, spec_vals_labs_names = __BEPSVals(udvs_mat, spec_inds,
                                                                                                bin_freqs, bin_wfm_type,
                                                                                                parm_dict, udvs_labs,
                                                                                                udvs_units)
        mode = parm_dict['VS_mode']
        ds_spec_inds_mat = create_spec_inds_from_vals(ds_spec_val_mat)

        # Make sure that the frequencies reset properly for user defined case
        spec_start = 0
        if mode == 'load user defined VS Wave from file':
            wave_form = udvs_mat[:, 3]
            for wave in wave_form:
                wave_freqs = bin_freqs[np.argwhere(bin_wfm_type==wave)].squeeze()
                num_bins = wave_freqs.size
                ds_spec_inds_mat[0, spec_start:spec_start+num_bins] = range(num_bins)
                spec_start += num_bins

    else:
        warn('Unknown format! Cannot generate Spectroscopic Values!')
        ds_spec_val_mat = []
        ds_spec_inds_mat = []
        ds_spec_val_labs = []
        ds_spec_val_units = []
        spec_vals_labs_names = []
    
    return ds_spec_val_mat, ds_spec_inds_mat, ds_spec_val_labs, ds_spec_val_units, spec_vals_labs_names


"""
BEHistogram Class and Functions
"""
class BEHistogram():
    # TODO: Turn into proper class
    # TODO: Parallelize Histogram generation
    """
    Class just functions as a container so we can have shared objects
    Chris Smith -- csmith55@utk.edu
    """
    def addBEHist(self, h5_path, max_mem_mb=1024, show_plot=True, save_plot=True):
        """
        This function adds Histgrams from the Main Data to the Plot Groups for
        an existing hdf5 BEPS datafile.

        Parameters
        ----------
        h5_path : string
            the path to the hdf5 datafile
        max_mem_mb : unsigned integer
            the maximum amount of memory to use during the binning
        show_plot : Boolean
            Should plot of the histograms be drawn after they are
            created
        save_plot : Boolean
            Should plots of the histograms be saved

        Returns
        -------
        None
        """
        hdf = ioHDF5(h5_path)
        h5_file = hdf.file

        print('Adding Histograms to file {}'.format(h5_file.name))
        print('Path to HDF5 file is {}'.format(hdf.path))

        max_mem = min(max_mem_mb*1024**2,0.75*getAvailableMem())

        h5_main = getDataSet(h5_file, 'Raw_Data')
        h5_udvs = getDataSet(h5_file,'UDVS')

        m_groups = [data.parent for data in h5_main]
        print('{} Measurement groups found.'.format(len(m_groups)))

        
        for im, group in enumerate(m_groups):

            p_groups = []

            mspecs = getDataSet(group,'Mean_Spectrogram')
            p_groups.extend([mspec.parent for mspec in mspecs])

            print('{} Plot groups in {}'.format(len(p_groups),group.name))

            for ip, p_group in enumerate(p_groups):
                try:
                    max_resp = getDataSet(group,'Max_Response')
                    min_resp = getDataSet(group,'Min_Response')
                except:
                    warn('Maximum and Minimum Response vectors not found for {}.'.format(p_group.name))
                    max_resp = []
                    min_resp = []

                print('Creating BEHistogram for Plot Group {}'.format(p_group.name))
                udvs_lab = p_group.attrs['Name']
                udvs_col = h5_udvs[im][h5_udvs[im].attrs[udvs_lab]]
                actual_udvs_steps = np.where(np.isnan(udvs_col)==False)[0]

                """
                Add the BEHistogram for the current plot group
                """
                plot_grp = MicroDataGroup(p_group.name.split('/')[-1], group.name[1:])
                plot_grp.attrs['Name'] = udvs_lab
                hist = BEHistogram()
                hist_mat, hist_labels, hist_indices, hist_indices_labels = hist.buildPlotGroupHist(h5_main[im], actual_udvs_steps, max_response=max_resp, min_response=min_resp, max_mem_mb=max_mem)
                ds_hist = MicroDataset('Histograms',hist_mat, dtype=np.int32, chunking=(1,hist_mat.shape[1]),compression='gzip')
                hist_slice_dict = dict()
                for hist_ind, hist_dim in enumerate(hist_labels):
                    hist_slice_dict[hist_dim] = (slice(hist_ind,hist_ind+1), slice(None))
                ds_hist.attrs['labels'] = hist_slice_dict
                ds_hist_indices = MicroDataset('Histograms_Indices',hist_indices,dtype=np.int32)
                hist_ind_dict = dict()
                for hist_ind_ind, hist_ind_dim in enumerate(hist_indices_labels):
                    hist_ind_dict[hist_ind_dim] = (slice(hist_ind_ind, hist_ind_ind+1),slice(None))
                ds_hist_indices.attrs['labels'] = hist_ind_dict
                ds_hist_labels = MicroDataset('Histograms_Labels',np.array(hist_labels))
                plot_grp.addChildren([ds_hist, ds_hist_indices, ds_hist_labels])
                hdf.writeData(plot_grp)

                if show_plot or save_plot:
                    if save_plot:
                        basename,junk = path.splitext(h5_path)
                        plotfile = '{}_MG{}_PG{}_Histograms.png'.format(basename,im,ip)
                    plot_histgrams(hist_mat, hist_indices, p_group, plotfile)
                    if show_plot:
                        plt.show()

        hdf.close()


    def buildBEHist(self,h5_main, max_response=[], min_response=[],max_mem_mb=1024, max_bins=256, debug=False):
        """
        Creates Histograms from dataset

        Parameters
        ----------
            h5_path : hdf5 reference to Main_Dataset

        Outputs:

        """

        free_mem = getAvailableMem()
        if debug: print('We have {} bytes of memory available'.format(free_mem))
        self.max_mem = min(max_mem_mb*1024**2,0.75*free_mem)

        """
        Check that max_response and min_response have been defined.
        Call __getminmaxresponse__ is not
        """
        if max_response == [] or min_response == []:
            max_response = np.max(np.abs(h5_main),axis=1)
            min_response = np.min(np.abs(h5_main),axis=1)


        self.max_response = np.mean(max_response)+3*np.std(max_response)
        self.min_response = np.max([0,np.mean(min_response)-3*np.std(min_response)])

        """
        Loop over all datasets
        """
        active_udvs_steps = getActiveUDVSsteps(h5_main) # technically needs to be done only once
        self.num_udvs_steps = len(active_udvs_steps)

        """
        Load auxilary datasets and extract needed parameters
        """
        spec_ind_mat = getAuxData(h5_main,auxDataName=['Spectroscopic_Indices'])[0]
        self.N_spectral_steps = np.shape(spec_ind_mat)[0]

        """
        Set up frequency axis of histogram, same for all histograms in a single dataset
        """
        freqs_mat = getAuxData(h5_main,auxDataName=['Bin_Frequencies'])[0]
        x_hist = np.array(spec_ind_mat)

        self.N_bins = np.size(freqs_mat)
        self.N_freqs = np.size(np.unique(freqs_mat))
        # print('There are {} total frequencies in this dataset'.format(self.N_bins))
        del freqs_mat, spec_ind_mat

        self.N_pixels = np.shape(h5_main)[1]
        # print('There are {} pixels in this dataset'.format(self.N_pixels))

        self.N_y_bins = np.int(np.min( (max_bins, np.rint(np.sqrt(self.N_pixels*self.N_spectral_steps)))))
#         self.N_y_bins = np.min( (max_bins, np.rint(2*(self.N_pixels*self.N_spectral_steps)**(1.0/3.0))))
        # print('{} bins will be used'.format(self.N_y_bins))

        ds_hist = self.__datasetHist(h5_main, active_udvs_steps, x_hist, debug)

        return ds_hist

    def buildPlotGroupHist(self, h5_main, active_spec_steps, max_response=[],
                           min_response=[], max_mem_mb=1024, max_bins=256,
                           std_mult=3, debug=False):
        """
        Creates Histograms for a given plot group

        Parameters
        ----------
        h5_main : HDF5 Dataset object
            Dataset to be historammed
        activ_spec_steps : numpy array
            active spectral steps in the current plot group
        max_response : numpy array
            maximum amplitude at each pixel
        min_response : numpy array
            minimum amplitude at each pixel
        max_mem : Unsigned integer
            maximum number of Mb allowed for use.  Used to calculate the
            number of pixels to load in a chunk
        max_bins : integer
            maximum number of spectroscopic bins
        std_mult : integer
            number of standard deviations from the mean of
            max_response and min_response to include in
            binning
        debug : boolean
            Turns on debug printing statements if true.  Default False.

        Returns
        -------
        hist_mat : 2d numpy array
            4 histograms as 1d arrays
        hist_labels : list of strings
            names for the 4 rows in hist_mat
        hist_indices : 2d numpy array
            the frequency and spectroscopic bins of each column in hist_mat
        hist_index_labels : list of strings
            labels for the hist_indices array

        """
        free_mem = getAvailableMem()
        if debug: print('We have {} bytes of memory available'.format(free_mem))
        self.max_mem = min(max_mem_mb,0.75*free_mem)

        """
        Check that max_response and min_response have been defined.
        Call __getminmaxresponse__ is not
        """
        if max_response == [] or min_response == []:
            max_response = np.amax(np.abs(h5_main),axis=0)
            min_response = np.amin(np.abs(h5_main),axis=0)

        self.max_response = np.mean(max_response)+std_mult*np.std(max_response)
        self.min_response = np.mean(min_response)-std_mult*np.std(min_response)
        del max_response,min_response

        """
        Load auxilary datasets and extract needed parameters
        """
        step_ind_mat = getAuxData(h5_main,auxDataName=['UDVS_Indices'])[0].value
        spec_ind_mat = getAuxData(h5_main,auxDataName=['Spectroscopic_Indices'])[0].value
        self.N_spectral_steps = np.size(step_ind_mat)

        active_udvs_steps = np.unique(step_ind_mat[active_spec_steps])
        self.num_udvs_steps = len(active_udvs_steps)

        """
        Set up frequency axis of histogram, same for all histograms in a single dataset
        """
        freqs_mat = getAuxData(h5_main,auxDataName=['Bin_Frequencies'])[0]
        x_hist = np.array([spec_ind_mat[0],step_ind_mat], dtype=np.int32)

        self.N_bins = np.size(freqs_mat)
        self.N_freqs = np.size(np.unique(freqs_mat))

        del freqs_mat, step_ind_mat, spec_ind_mat

        self.N_pixels = np.shape(h5_main)[0]

#         self.N_y_bins = np.int(np.min( (max_bins, np.rint(np.sqrt(self.N_pixels*self.N_spectral_steps)))))
        self.N_y_bins = np.int(np.min( (max_bins, np.rint(2*(self.N_pixels*self.N_spectral_steps)**(1.0/3.0)))))


        ds_hist = self.__datasetHist(h5_main, active_udvs_steps, x_hist, debug)
        if debug: print(np.shape(ds_hist))
        if debug: print('ds_hist max',np.max(ds_hist),
                        'ds_hist min',np.min(ds_hist))

        hist_mat, hist_labels, hist_indices, hist_index_labels = self.__reshapeHist(ds_hist)

        return hist_mat, hist_labels, hist_indices, hist_index_labels

    def __reshapeHist(self,ds_hist):
        """
        Reshape the histogram matrix into table, and build the associated index table

        Parameters
        ----------
        ds_hist : numpy array
            the 4 histogram matrices

        Returns
        -------
        hist_mat : 2d numpy array
            the 4 histograms as 1d arrays
        hist_labels : list of strings
            names for the 4 rows in hist_mat
        hist_indices : 2d numpy array
            the frequency and spectroscopic bins of each column in hist_mat
        hist_index_labels : list of strings
            labels for the hist_indices array
        """
        hist_shape = ds_hist.shape

        hist_mat = np.reshape(ds_hist, (hist_shape[0],hist_shape[1]*hist_shape[2]))

        hist_labels = ['Amplitude','Phase','Real Part','Imaginary Part']

        hist_indices = np.zeros((2,hist_mat.shape[1]), dtype=np.int32)

        hist_index_labels = ['Frequency Bin','Spectroscopic Bin']

        for isbin in range(hist_shape[1]):
            for ifbin in range(hist_shape[2]):
                ihbin = ifbin+isbin*hist_shape[2]
                hist_indices[0,ihbin] = ifbin
                hist_indices[1,ihbin] = isbin

        return hist_mat, hist_labels, hist_indices, hist_index_labels

    def __datasetHist(self, h5_main, active_udvs_steps, x_hist, debug=False):
        """
        Create the histogram for a single dataset

        Parameters
        ----------
        h5_main : HDF5 Dataset
            Main_Dataset to be histogramed
        active_udvs_steps : numpy array
            the active udvs steps in the current plot group
        x_hist : 1d numpy array
            the spectroscopic indices matrix, used to find the
            spectroscopic indices of each udvs step

        Returns
        -------
        ds_hist : numpy array
            the 4 histogram matrices

        """


        """
        Estimate maximum number of pixels to read at once
        """
        max_pixels = maxReadPixels(self.max_mem,self.N_pixels, self.num_udvs_steps, bytes_per_bin=h5_main.dtype.itemsize*self.N_y_bins*self.N_freqs)

        """
        Divide the pixels into chunks that will fit in memory
        """
        pix_chunks = np.append(np.arange(0,self.N_pixels,max_pixels,dtype=np.int),self.N_pixels)

        """
        Initialize the histograms
        """
        ds_hist = np.zeros((4,self.N_freqs,self.N_y_bins),dtype=np.int32)

        """
        loop over pixels
        """
        for ichunk in range(len(pix_chunks)-1):
            if debug: print('pixel chunk', ichunk)

            chunk = range(pix_chunks[ichunk], pix_chunks[ichunk+1])

            """
        Loop over active UDVS steps
            """
            for iudvs in range(self.num_udvs_steps):
                selected = (iudvs+chunk[0]*self.num_udvs_steps) % np.rint(self.num_udvs_steps*self.N_pixels/10) == 0
                if selected:
                    per_done = np.rint(100*(iudvs+chunk[0]*self.num_udvs_steps)/(self.num_udvs_steps*self.N_pixels))
                    print('Binning BEHistogram...{}% --pixels {}-{}, step # {}'.format(per_done,chunk[0],chunk[-1],iudvs))
                udvs_step = active_udvs_steps[iudvs]
                if debug: print('udvs step', udvs_step)

                """
        Get the correct Spectroscopic bins for the current UDVS step
        Read desired pixel chunk from these bins for Main_Data into data_mat
                """
                udvs_bins = np.where(x_hist[1] == udvs_step)[0]
                if debug:
                    print(np.shape(x_hist))
                data_mat = h5_main[pix_chunks[ichunk]:pix_chunks[ichunk+1], (udvs_bins)]

                """
        Get the frequecies that correspond to the current UDVS bins from the total x_hist
                """
                this_x_hist = np.take(x_hist[0], udvs_bins)
                this_x_hist = this_x_hist-this_x_hist[0]
                this_x_hist = np.transpose(np.tile(this_x_hist, (1, pix_chunks[ichunk+1]-pix_chunks[ichunk])))
                this_x_hist = np.squeeze(this_x_hist)

                N_x_bins = np.shape(this_x_hist)[0]
                if debug:
                    print('N_x_bins',N_x_bins)
                    print(this_x_hist)
                    print(np.shape(this_x_hist))
                """
        Create weighting vector.  If setting all to one value, can be a scalar.
                """
                weighting_vec = 1

                if debug: print(np.shape(data_mat))

                """
        Set up the list of functions to call and their corresponding maxima and minima
                """
                func_list = [np.abs, np.angle, np.real, np.imag]
                max_list = [self.max_response, np.pi, self.max_response, self.max_response]
                min_list = [self.min_response, -np.pi, self.min_response, self.min_response]
                """
        Get the Histograms and store in correct place in ds_hist
                """
                for ifunc, func in enumerate(func_list):
                    chunk_hist = buildHistogram(this_x_hist,
                                                data_mat,
                                                N_x_bins,
                                                self.N_y_bins,
                                                weighting_vec,
                                                min_list[ifunc],
                                                max_list[ifunc],
                                                func,
                                                debug)
                    if debug:
                        print('chunkhist-{}'.format(func.__name__), np.shape(chunk_hist))
                        print(chunk_hist.dtype)

                    for (i, ifreq) in enumerate(udvs_bins):
                        ids_freq = this_x_hist[i]
                        if debug:
                            print(i, ifreq)
                            print(ids_freq)
                        ds_hist[ifunc, ids_freq, :] = np.add(ds_hist[ifunc, ids_freq, :], chunk_hist[i, :])

        return ds_hist

    


"""
Created on 7/17/16 10:08 AM
@author: Numan Laanait, Suhas Somnath, Chris Smith
"""

from __future__ import division, print_function, absolute_import, unicode_literals
from warnings import warn

import numpy as np
import psutil
import scipy
from .guess_methods import GuessMethods
from .fit_methods import Fit_Methods
from ..io.hdf_utils import checkIfMain, getAuxData
from ..io.io_hdf5 import ioHDF5
from ..io.io_utils import getAvailableMem, recommendCores
from .optimize import Optimize


class Model(object):
    """
    Encapsulates the typical routines performed during model-dependent analysis of data.
    This abstract class should be extended to cover different types of imaging modalities.

    Parameters
    ----------
    h5_main : h5py.Dataset instance
        The dataset over which the analysis will be performed. This dataset should be linked to the spectroscopic
        indices and values, and position indices and values datasets.
    variables : list(string), Default ['Frequency']
        Lists of attributes that h5_main should possess so that it may be analyzed by Model.
    parallel : bool, optional
        Should the parallel implementation of the fitting be used.  Default True.

    Returns
    -------
    None

    """
    def __init__(self, h5_main, variables=['Frequency'], parallel=True):
        """
        For now, we assume that the guess dataset has not been generated for this dataset but we will relax this requirement
        after testing the basic components.

        """
        # Checking if dataset is "Main"
        if self._is_legal(h5_main, variables):
            self.h5_main = h5_main
            self.hdf = ioHDF5(self.h5_main.file)

        else:
            raise ValueError('Provided dataset is not a "Main" dataset with necessary ancillary datasets')

        # Checking if parallel processing will be used
        self._parallel = parallel

        # Determining the max size of the data that can be put into memory
        self._set_memory_and_cores()

        self._start_pos = 0
        self._end_pos = self.h5_main.shape[0]
        self.h5_guess = None
        self.h5_fit = None

        self.data = None
        self.guess = None
        self.fit = None

    def _set_memory_and_cores(self, verbose=False):
        """
        Checks hardware limitations such as memory, # cpus and sets the recommended datachunk sizes and the
        number of cores to be used by analysis methods.

        Parameters
        ----------
        verbose : Boolean (Optional)
            Whether or not to print log statements

        Returns
        -------
        None

        """

        if self._parallel:
            self._maxCpus = max(1, psutil.cpu_count() - 2)
        else:
            self._maxCpus = 1

        if self._maxCpus == 1:
            self._parallel = False

        self._maxMemoryMB = getAvailableMem() / 1024**2 # in Mb

        self._maxDataChunk = int(self._maxMemoryMB / self._maxCpus)

        # Now calculate the number of positions that can be stored in memory in one go.
        mb_per_position = self.h5_main.dtype.itemsize * self.h5_main.shape[1] / 1024.0 ** 2
        self._max_pos_per_read = int(np.floor(self._maxDataChunk / mb_per_position))
        if verbose:
            print('Allowed to read {} pixels per chunk'.format(self._max_pos_per_read))

    def _is_legal(self, h5_main, variables):
        """
        Checks whether or not the provided object can be analyzed by this Model class.
        Classes that extend this class will do additional checks to ensure that the supplied dataset is legal.

        Parameters
        ----
        h5_main : h5py.Dataset instance
            The dataset over which the analysis will be performed. This dataset should be linked to the spectroscopic
            indices and values, and position indices and values datasets.

        variables : list(string)
            The dimensions needed to be present in the attributes of h5_main to analyze the data with Model.

        Returns
        -------
        legal : Boolean
            Whether or not this dataset satisfies the necessary conditions for analysis

        """

        # Check if h5_main is a "Main" dataset
        cond_A = checkIfMain(h5_main)

        # Check if variables are in the attributes of spectroscopic indices
        h5_spec_vals = getAuxData(h5_main, auxDataName=['Spectroscopic_Values'])[0]
        # assert isinstance(h5_spec_vals, list)
        cond_B = set(variables).issubset(set(h5_spec_vals.attrs.keys()))

        if cond_A and cond_B:
            legal = True
        else:
            legal = False

        return legal

    def _get_data_chunk(self, verbose=False):
        """
        Returns the next chunk of data for the guess or the fit

        Parameters
        -----
        None

        Returns
        --------

        """
        if self._start_pos < self.h5_main.shape[0]:
            self._end_pos = int(min(self.h5_main.shape[0], self._start_pos + self._max_pos_per_read))
            self.data = self.h5_main[self._start_pos:self._end_pos, :]
            if verbose:
                print('Reading pixels {} to {} of {}'.format(self._start_pos, self._end_pos, self.h5_main.shape[0]))

            # Now update the start position
            self._start_pos = self._end_pos
        else:
            if verbose:
                print('Finished reading all data!')
            self.data = None


    def _get_guess_chunk(self):
        """
        Returns a chunk of guess dataset corresponding to the main dataset.
        Should be called BEFORE _get_data_chunk since it relies upon current values of
        `self._start_pos`, `self._end_pos`

        Parameters
        -----
        None

        Returns
        --------

        """
        if self.data is None:
            self._end_pos = int(min(self.h5_main.shape[0], self._start_pos + self._max_pos_per_read))
            self.guess = self.h5_guess[self._start_pos:self._end_pos, :]
        else:
            self.guess = self.h5_guess[self._start_pos:self._end_pos, :]

    def _set_results(self, is_guess=False):
        """
        Writes the provided guess or fit results into appropriate datasets.
        Given that the guess and fit datasets are relatively small, we should be able to hold them in memory just fine

        Parameters
        ---------
        is_guess : Boolean
            Flag that differentiates the guess from the fit

        """
        statement = 'guess'

        if is_guess:
            targ_dset = self.h5_guess
            source_dset = self.guess
        else:
            statement = 'fit'
            targ_dset = self.h5_fit
            source_dset = self.fit

        """print('Writing data to positions: {} to {}'.format(self.__start_pos, self._end_pos))
        targ_dset[self._start_pos:self._end_pos, :] = source_dset"""
        targ_dset[:, :] = source_dset

        # flush the file
        self.hdf.flush()
        print('Finished writing ' + statement + ' results to file!')

    def _create_guess_datasets(self):
        """
        Model specific call that will write the h5 group, guess dataset, corresponding spectroscopic datasets and also
        link the guess dataset to the spectroscopic datasets. It is recommended that the ancillary datasets be populated
        within this function.

        The guess dataset will NOT be populated here but will be populated by the __setData function
        The fit dataset should NOT be populated here unless the user calls the optimize function.

        Parameters
        --------
        None

        Returns
        -------
        None

        """
        warn('Please override the _create_guess_datasets specific to your model')
        self.guess = None  # replace with actual h5 dataset
        pass

    def _create_fit_datasets(self):
        """
        Model specific call that will write the h5 group, fit dataset, corresponding spectroscopic datasets and also
        link the fit dataset to the spectroscopic datasets. It is recommended that the ancillary datasets be populated
        within this function.

        The fit dataset will NOT be populated here but will be populated by the __setData function
        The guess dataset should NOT be populated here unless the user calls the optimize function.

        Parameters
        --------
        None

        Returns
        -------
        None

        """
        warn('Please override the _create_fit_datasets specific to your model')
        self.fit = None # replace with actual h5 dataset
        pass

    def do_guess(self, processors=None, strategy='wavelet_peaks',
                 options={"peak_widths": np.array([10, 200]), "peak_step":20}):
        """

        Parameters
        ----------
        strategy: string (optional)
            Default is 'Wavelet_Peaks'.
            Can be one of ['wavelet_peaks', 'relative_maximum', 'gaussian_processes'].
            For updated list, run GuessMethods.methods
        processors : int (optional)
            Number of cores to use for computing. Default = all available - 2 cores
        options: dict
            Default, options for wavelet_peaks {"peaks_widths": np.array([10,200]), "peak_step":20}.
            Dictionary of options passed to strategy. For more info see GuessMethods documentation.

        Returns
        -------

        """

        self._start_pos = 0

        processors = recommendCores(self._max_pos_per_read, processors)

        self._get_data_chunk()
        gm = GuessMethods()
        results = list()
        if strategy in gm.methods:
            print("Using %s to find guesses...\n" % strategy)
            while self.data is not None:
                opt = Optimize(data=self.data, parallel=self._parallel)
                temp = opt.computeGuess(processors=processors, strategy=strategy, options=options)
                results.append(self._reformat_results(temp, strategy))
                self._get_data_chunk()

            # reorder to get one numpy array out
            self.guess = np.hstack(tuple(results))
            print('Completed computing guess. Writing to file.')

            # Write to file
            self._set_results(is_guess=True)
        else:
            raise KeyError('Error: %s is not implemented in pycroscopy.analysis.GuessMethods to find guesses' %
                           strategy)

        return self.guess

    def _reformat_results(self, results, strategy='wavelet_peaks'):
        """
        Model specific restructuring / reformatting of the parallel compute results

        Parameters
        ----------
        results : array-like
            Results to be formatted for writing
        strategy : str
            The strategy used in the fit.  Determines how the results will be reformatted.
            Default 'wavelet_peaks'

        Returns
        -------
        results : numpy.ndarray
            Formatted array that is ready to be writen to the HDF5 file 

        """
        return np.array(results)

    def _create_fit_dataset(self):
        """
        Model specific call that will write the HDF5 fit dataset. pycroscopy requires that the h5 group, guess dataset,
        corresponding spectroscopic and position datasets be created and populated at this point.
        This function will create the HDF5 dataset for the fit and link it to same ancillary datasets as the guess.
        The fit dataset will NOT be populated here but will instead be populated using the __setData function
        """
        warn('Please override the _create_fit_dataset specific to your model')
        self.h5_fit = None  # replace with actual h5 dataset
        pass

    def do_fit(self, processors=None, solver_type='least_squares', solver_options={'jac': '2-point'},
               obj_func={'class': 'Fit_Methods', 'obj_func': 'SHO', 'xvals': np.array([])}):
        """
        Generates the fit for the given dataset and writes back to file

        Parameters
        ----------
        processors : int
            Number of cpu cores the user wishes to run on.  The minimum of this and self._maxCpus is used.
        solver_type : str
            The name of the solver in scipy.optimize to use for the fit
        solver_options : dict
            Dictionary of parameters to pass to the solver specified by `solver_type`
        obj_func : dict
            Dictionary defining the class and method containing the function to be fit as well as any 
            additional function parameters.

        Returns
        -------

        """
        if self.h5_guess is None:
            print("You need to guess before fitting\n")
            return None

        processors = recommendCores(self._max_pos_per_read, processors)

        self._start_pos = 0
        self._get_guess_chunk()
        self._get_data_chunk()
        results = list()
        legit_solver = solver_type in scipy.optimize.__dict__.keys()
        legit_obj_func = obj_func['obj_func'] in Fit_Methods().methods
        if legit_solver and legit_obj_func:
            print("Using solver %s and objective function %s to fit your data\n" % (solver_type, obj_func['obj_func']))
            while self.data is not None:
                opt = Optimize(data=self.data, guess=self.guess, parallel=self._parallel)
                temp = opt.computeFit(processors=processors, solver_type=solver_type, solver_options=solver_options,
                                      obj_func=obj_func)
                # TODO: need a different .reformatResults to process fitting results
                results.append(self._reformat_results(temp, obj_func['obj_func']))
                self._get_guess_chunk()
                self._get_data_chunk()

            self.fit = np.hstack(tuple(results))
            self._set_results()

        elif legit_obj_func:
            raise KeyError('Error: Solver "%s" does not exist!. For additional info see scipy.optimize\n' % solver_type)

        elif legit_solver:
            raise KeyError('Error: Objective Functions "%s" is not implemented in pycroscopy.analysis.Fit_Methods' %
                           obj_func['obj_func'])

        return results
